// vincia24.cc is a part of the VINCIA event generator.
// Copyright (C) 2016 Peter Skands.
// VINCIA is licenced under the GNU GPL version 2, see COPYING for details.
// Please respect the MCnet Guidelines, see GUIDELINES for details.

// This is a simple test program to illustrate how several HepMC
// files can be used for uncertainty bands

#include "Vincia/Vincia.h"
#include "Pythia8Plugins/HepMC2.h"

using namespace Pythia8;
using namespace Vincia;

int main( int argc, char* argv[] ) {
  
  // Weights, modify weightVec to include the desired 
  // uncertainty variations
  vector<int> weightVec;
  weightVec.push_back( 0); // central weight
  //weightVec.push_back( 1); // variations
  weightVec.push_back( 2);
  weightVec.push_back( 3);
  //weightVec.push_back( 4);
  //weightVec.push_back( 5);
  //weightVec.push_back( 6);
  //weightVec.push_back( 7);
  //weightVec.push_back( 8);
  //weightVec.push_back( 9);
  //weightVec.push_back(10);
  int nWeights = (int)weightVec.size();
  
  // Additional HepMC file which stores the weights of all
  // uncertainty variations
  bool addAllWeights = true;
  
  // Check that correct number of command-line arguments
  if (argc != 3) {
    // Central weight present?
    bool central     = ( find(weightVec.begin(), weightVec.end(), 0) != weightVec.end() );
    // List of weights without central
    string HepMClist = "HepMCprefix{";
    for (int i=0; i<nWeights; i++) if (weightVec[i] != 0) {
      HepMClist += static_cast<ostringstream*>( &(ostringstream() << weightVec[i]) )->str();;
      if (i != nWeights-1) HepMClist += ",";
    }
    HepMClist += "}";
    // List of weights including central
    string HepMCcomp = HepMClist;
    if (central) HepMCcomp.insert(12,"0,");
    cout << endl
         << " Unexpected number of command-line arguments." << endl << endl
         << " Use the program like this: ./vincia24 vincia24.cmnd HepMCprefix"
         << endl << endl
         << " The program will create the HepMC file(s)";
    if (central) cout << " HepMCprefix0 for the central weight and" << endl;
    cout << " " << HepMClist << " for the uncertainy variations." << endl
         << " Variations can be added/removed in weightVec in the program." << endl;
    if (addAllWeights)
      cout << " In addition the file HepMCprefix will store a vector of weights. " << endl
           << " Can be switched off in the program with addAllWeights." << endl;
    cout << endl
         << " For passing events directly to Rivet with fifo pipes do: " << endl;
    if (addAllWeights) cout << "   mkfifo HepMCprefix " << HepMCcomp << endl;
    else cout << "   mkfifo " << HepMCcomp << endl;
    cout << "   ./vincia24 vincia24.cmnd HepMCprefix &" << endl
         << "   rivet -H YODA_NAME.yoda -a ANALYSIS_NAME HepMCprefixX " << endl
         << " to analyse the Xth uncertainy variations." << endl
         << " Note that you have to execture the rivet command on all fifos." << endl
         << endl;
    return 1;
  }
  
  // Pythia and Vincia
  Pythia pythia;
  VinciaPlugin vincia(&pythia, argv[1]);
  vincia.init();
  
  // Settings and Info pointer
  Settings& settings = pythia.settings;
  Info& info         = pythia.info;
  
  // Check that uncertainty bands are switched on
  if (!settings.flag("Vincia:uncertaintyBands")) {
    cout << endl
         << " Uncertainty bands are not switched on. Exiting."
         << endl;
    return 1;
  }
  
  // Weight sum, used to calculate the modified cross section
  // (for all hard processes not including QCD jets in the final
  // state, note that decay products, e.g. Z->qqbar, no not count
  // as QCD jets in this cas)
  double weightSum = 0.0;
  
  // HepMC
  string hepmcFile = string(argv[2]);  
  HepMC::Pythia8ToHepMC ToHepMC;
  // Switch off warnings for parton-level events.
  ToHepMC.set_print_inconsistency(false);
  // Vector of events, one for each vector of weights
  vector< HepMC::IO_GenEvent* > HepMCevent;
  HepMCevent.resize(addAllWeights ? nWeights+1 : nWeights);
  for (int iVar=0; iVar<nWeights; iVar++) {
    string name = hepmcFile +
                  static_cast<ostringstream*>( &(ostringstream() << 
                                               weightVec[iVar]) )->str();;
    HepMCevent[iVar] = new HepMC::IO_GenEvent(name, std::ios::out);
  }
  if (addAllWeights)
    HepMCevent[nWeights] = new HepMC::IO_GenEvent(hepmcFile, std::ios::out);
  
  // Number of events
  int nEvent = settings.mode("Main:numberOfEvents");

  // Begin event loop
  for (int iEvent = 1; iEvent <= nEvent; ++iEvent) {
    
    if (!pythia.next()) continue;
    
    // Full weight (includes reweighting for MC violations, if 
    // Vincia:allowReweighting is switched on)
    double weight     = vincia.weight();
    // Initial weight, use for cross section
    double weightInit = vincia.weightInit();
    // Pythia's weight is already accounted for in the CS
    weightSum        += weightInit/info.weight();
    
    // Create vector with weights
    vector<double> weights;
    weights.push_back(weight);
    for (int iVar=1; iVar<11; iVar++)
      weights.push_back(vincia.weight(iVar)); 
    
    // Cross section information
    double cs    = info.sigmaGen()*weightSum/((double)(iEvent)) * 1e9;
    double cserr = info.sigmaErr()*weightSum/((double)(iEvent)) * 1e9;
    
    // Fill HepMC files
    HepMC::GenEvent* HepMCeventNow = new HepMC::GenEvent();
    ToHepMC.fill_next_event( pythia, HepMCeventNow );
    // Set the cross section
    HepMCeventNow->cross_section()->set_cross_section( cs, cserr );
    // Set the weights
    for (int iVar=0; iVar<nWeights; iVar++) {
      HepMCeventNow->weights()[0] = weights[weightVec[iVar]];
      *HepMCevent[iVar] << HepMCeventNow;
    }
    if (addAllWeights) {
      HepMCeventNow->weights() = weights;
      *HepMCevent[nWeights] << HepMCeventNow;
    }
    // Set vector of weights
    delete HepMCeventNow;
        
  } // End of event loop
    
  // Print cross section
  cout << endl << scientific  << " Cross Section: " 
       << info.sigmaGen()*weightSum/((double)(nEvent)) << endl << endl;
  cout.precision(3); 
  cout << fixed;
  // Print statistics and shower infos
  vincia.fsrShower.printInfo();
  vincia.isrShower.printInfo();
  pythia.stat();
  
  // Done
  return 0;
  
}
