// VinciaMEcorrections.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains header information for VINCIA's MEcorrections class.
// Added by NF. Includes ISR clusterings and does not know about the Born
// kinematics.
  
#ifndef VINCIAMECORRECTIONS_H
#define VINCIAMECORRECTIONS_H

#include "Pythia8/Basics.h"
#include "Pythia8/Event.h"
#include "Vincia/VinciaCommon.h"

#include "Vincia/AntennaSets.h"
#include "Vincia/Resolution.h"
#include "Vincia/SubtractionTerms.h"
#include "Vincia/MatchingGKS.h"

namespace Vincia {
  
  static vector< vector< vector<double> > > DEF_VEC;
  
  const int procClassMin         = 1;
  const int procClassMax         = 3;
  
  // Process classes for which we can do MECs
  const int singleBosonDecay     = 1;
  const int singleBosonProd      = 2;  
  const int pureQCD              = 3;
  
  // Max number of particles for which we can do MECs
  const int nMax[] = { 0, 7, 6, 6 };
  // 0 for unknown process
  // 7 for single boson decays, up to X -> 6
  // 6 for single boson production, up to PP -> X+3
  // 6 for QCD, up to PP -> 4

  class MEcorrections {
    
  public:
    
    // Constructor.
    MEcorrections() : isInitPtr(false), isInit(false) {};

    // Destructor.
    virtual ~MEcorrections() {};

    // Initialize Pointers
    void initPtr(Settings* settingsPtrIn, MG4interface* mgInterfacePtrIn,
                 Rndm* rndmPtrIn, Resolution* resolutionPtrIn,
		 VinciaCommon* vinComPtrIn, VinClu* vinCluPtrIn) {
      settingsPtr    = settingsPtrIn;
      mgInterfacePtr = mgInterfacePtrIn;
      rndmPtr        = rndmPtrIn;
      resolutionPtr  = resolutionPtrIn;
      vinComPtr      = vinComPtrIn;
      vinCluPtr      = vinCluPtrIn;
      aSptr          = &vinComPtr->alphaStrong;
      aSCMWptr       = &vinComPtr->alphaStrongCMW;
      aSdefPtr       = &vinComPtr->alphaStrongDef;
      aSCMWdefPtr    = &vinComPtr->alphaStrongDefCMW;
      isInitPtr      = true;
    }
    
    // Initialize
    void init();
    
    // Wrapper function to return a specific antenna inside a specific antenna set
    // 0: User, 1: Def, 2: Max, 3: Min    
    AntennaFSR* getAntFSR(int iAnt, int iSet=0) {
      return antSetFSR[iSet]->getAnt(iAnt);
    }
    AntennaISR* getAntISR(int iAnt, int iSet=0) {
      return antSetISR[iSet]->getAnt(iAnt);
    }
    
    // Initialize Pointers to Antenna sets
    void initAntPtr(AntennaSetFSR* antFSRusr, AntennaSetISR* antISRusr,
                    AntennaSetFSR* antFSRdef=0, AntennaSetISR* antISRdef=0,
                    AntennaSetFSR* antFSRmax=0, AntennaSetISR* antISRmax=0,
                    AntennaSetFSR* antFSRmin=0, AntennaSetISR* antISRmin=0) {
      antSetFSR.resize(0);
      antSetFSR.push_back(antFSRusr);
      if (antFSRdef != 0 && antFSRmax != 0 && antFSRmin != 0) {
        antSetFSR.push_back(antFSRdef);
        antSetFSR.push_back(antFSRmax);
        antSetFSR.push_back(antFSRmin);
      }
      antSetISR.resize(0);
      antSetISR.push_back(antISRusr);
      if (antISRdef != 0 && antISRmax != 0 && antISRmin != 0) {
        antSetISR.push_back(antISRdef);
        antSetISR.push_back(antISRmax);
        antSetISR.push_back(antISRmin);
      }
    }
    
    // Can we do the process
    bool canDoMEC(vector<Particle> parts, bool hardProcess) {
      if (procClass < procClassMin || procClass > procClassMax) {
        if (verbose >= 4) {
          if (hardProcess) printOut("Vincia::canDoMEC", 
                                    "Requested MEC for unknown process class in hard process!");
          else printOut("Vincia::canDoMEC", 
                        "Requested MEC for unknown process class!");
        }
        return false;
      }
      int nParticles = (int)parts.size();
      if (nParticles > nMax[procClass]) {
        if (verbose >= 4) {
          if (hardProcess) printOut("Vincia::canDoMEC", 
                                    "Requested MEC order too large in hard process!");
          else printOut("Vincia::canDoMEC", 
                        "Requested MEC order too large!");
        }
        return false;
      }
      // Check single boson production / decay
      if (procClass == singleBosonDecay || procClass == singleBosonProd) {
        // We can do Z=23, W=24, H=25
        bool foundBoson = false;
        for (int i=0; i<nParticles; i++) {
          if (parts[i].idAbs()>22 && parts[i].idAbs()<26) {
            // Make sure it's only a single boson
            if (foundBoson) return false;
            foundBoson = true;
          }
        }
        return foundBoson;
      }
      // Check QCD
      if (procClass == pureQCD) {
        bool foundNonQCD = false;
        for (int i=0; i<nParticles; i++) {
          if (parts[i].idAbs()>6 && parts[i].idAbs()!=21)
            foundNonQCD = true;
        }
        return (!foundNonQCD);
      }
      // All good!
      return true;
    }
    
    // Set process class
    void setProcClass(int procClassIn) { procClass = procClassIn; }
    
    // Some helpers for process types
    bool decayME() { return (procClass == singleBosonDecay); }
    bool prodME()  { return (procClass == singleBosonProd ); }
    bool twoToN()  { return (procClass == singleBosonProd || procClass == pureQCD); }
    
    // Include Ariadne factor?
    void setPari(bool doPariIn) { doPari = doPariIn; }
    
    // Set minimal number of quark lines and
    // Higgs-g for Higgs-q coupling off?
    // 0 = interference on
    // 1 = interference off and qqb->H as basic process
    // 2 = interference off and  gg->H as basic process
    void setClustProps(int nQLminIn, int HinterIn) { 
      nQLmin = nQLminIn; 
           if (HinterIn == 1) mgInterfacePtr->setHinterferenceOff(true);
      else if (HinterIn == 2) mgInterfacePtr->setHinterferenceOff(false);
    }
    
    // Set CMW, to be used for hard jet
    void setCMW(bool CMWin) { doCMWnow = CMWin; }
  
    // Main Function: Get all possible histories and perform MEn+1 / Sum ant*MEn
    vector<double> MEcorrection(vector<Particle> parts);
    
    // Functions for flat phase space scan
    void PSME1stOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                          vector<int>& nrOrdPaths, bool print, 
                          bool isQCD = false);
    
    void PSME2ndOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                          vector< vector<double> >& Q2clust1, vector< vector<double> >& Q2clust2,
                          vector<int>& nrOrdPaths, bool print, 
                          bool isQCD = false, vector< vector< vector<double> > >& Q2clust2QCD = DEF_VEC,
                          vector< vector< vector<double> > >& Q2fac = DEF_VEC);
    
    void PSME3rdOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                          vector< vector<double> >& Q2clust1,
                          vector<int>& nrOrdPaths, bool print);
    
    void PSME4thOrderFlat(vector<Particle> parts, vector<double>& PSoverME, 
                          vector< vector<double> >& Q2clust1,
                          vector<int>& nrOrdPaths, bool print);
    
    // Perform the clustering
    void clustering(vector<Particle> parts);
    
    // Get clustered momenta
    vector<Particle> clustII(Particle inA, Particle inB, Particle inJ, 
                             int iA, int iB, int iJ,
                             vector<Particle> partonsIn, int iAnt, bool swapped, 
                             vector<double>* invs);
    
    vector<Particle> clustIF(Particle inA, Particle inK, Particle inJ,
                             int iA, int iK, int iJ,
                             vector<Particle> partonsIn, int iAnt, 
                             vector<double>* invs);
    
    vector<Particle> clustFF(Particle inI, Particle inK, Particle inJ,
                             int iI, int iK, int iJ,
                             vector<Particle> partonsIn, int* iAnt,
                             vector<double>* invs);
    
    // Get the order of the n partons for evaluating the matrix elements
    vector< vector<int> > orderForME(vector<Particle> parts);
    
    // Evaluate a matrix element for the defined process classes
    double evalME(vector<Particle> parts);
  
    // Function for the unordering: Finds the smallest scale and a list
    // of partons that should be unordered
    void unorderingScale(vector<Particle> partonsIn, double* Q2min, 
                         vector<int>* unOrdPartons);
    // Unordering scale for uncertainty bands (with different evolution types)
    double unorderingScaleUB(vector<Particle> partonsIn, vector<double> evolTypes);
    
    // Find the colour chains
    vector< vector<int> > getColChains(vector<Particle> partonsIn);
    
    // Check if given quark pair can be clustered (ie not the same chain)
    bool canClustQQB(vector<Particle> partonsIn, vector< vector<int> > colChains, 
                     int iJ, int iPQ);
    
    // Which evolution type?
    vector<double> getEvolTypes() {
      vector<double> evolTypes;
      evolTypes.push_back(evolTypeGemiFF);
      evolTypes.push_back(evolTypeGsplitFF);
      evolTypes.push_back(evolTypeGemiI);
      evolTypes.push_back(evolTypeGsplitI);
      evolTypes.push_back(evolTypeGsplitF);
      evolTypes.push_back(evolTypeQconvI);
      return evolTypes;
    }
    void setEvolTypes(vector<double> evolTypes) {
      evolTypeGemiFF   = evolTypes[0];
      evolTypeGsplitFF = evolTypes[1];
      evolTypeGemiI    = evolTypes[2];
      evolTypeGsplitI  = evolTypes[3];
      evolTypeGsplitF  = evolTypes[4];
      evolTypeQconvI   = evolTypes[5];
    }
    
    // Some helper functions
    int countNrQuarks(vector<Particle> parts, int emi1 = -1, int emi2 = -1) {
      int countQ = 0;
      for (int i=0; i<(int)parts.size(); i++) {
        if ( (i != emi1) && (i != emi2) ) if (abs(parts[i].id())<6) countQ++;
      }
      return countQ;
    }
    int countNrGluons(vector<Particle> parts) {
      int countG = 0;
      for (int i=0; i<(int)parts.size(); i++)
        if (parts[i].id()==21) countG++;
      return countG;
    }
    int countNrPartons(vector<Particle> parts) {
      int countP = 0;
      for (int i=0; i<(int)parts.size(); i++)
        if ( (abs(parts[i].id())<6) || (parts[i].id()==21) ) countP++;
      return countP;
    }
    
    void setVerbose(int verboseIn) { verbose = verboseIn; }

  private:

    // Is initialized?
    bool isInitPtr, isInit;

    // Pointer to PYTHIA 8 Settings database and Random-number generator
    Settings* settingsPtr;
    Rndm*     rndmPtr;
    // MadGraph interface
    MG4interface* mgInterfacePtr;
    // Pointer to Evolution Variable
    Resolution* resolutionPtr;
    // Pointer to VINCIA common function classes
    VinciaCommon* vinComPtr;
    VinClu* vinCluPtr;
    
    // Pointer tos trong coupling in MSbar and CMW schemes
    AlphaStrong* aSptr;
    AlphaStrong* aSCMWptr;
    AlphaStrong* aSdefPtr;
    AlphaStrong* aSCMWdefPtr;
    
    // Antenna Sets
    vector<AntennaSetFSR*> antSetFSR;
    vector<AntennaSetISR*> antSetISR;
    
    bool doCMWnow;
    // User alphaS settings
    int CMWtypeII, CMWtypeIF, CMWtypeFF;
    double aSkMu2I, aSkMu2F;
    double aSmax, aSmuFreeze, aSmuMin;
    double mu2freeze, mu2min;
    // default alphaS settings
    int CMWtypeIIdef, CMWtypeIFdef, CMWtypeFFdef;
    double aSkMu2Idef, aSkMu2Fdef;
    double aSmaxDef, aSmuFreezeDef, aSmuMinDef;
    double mu2freezeDef, mu2minDef;

    // Functions used by NLO interface (disabled for now)
    //Functions* functionsPtr;
    
    // Matching settings
    bool meCorrections, matchingFullColor;
    int  matchingLO;
    
    // Uncertainties
    bool uncertaintyBands;
    int nWeights;
    double aSkMuVar2;
    
    // Ariadne correction factor?
    bool doPari;
    
    // Evolution type
    int evolTypeGemiFF, evolTypeGsplitFF;
    int evolTypeGemiI, evolTypeGsplitI, evolTypeGsplitF, evolTypeQconvI;
    
    // Ordering mode
    int orderingMode;
    
    // For QCD processes
    bool hardJetsQCD;
    int factorScale2;
    double factorMultFac, factorFixScale;
    
    // Process class
    int procClass;
    
    // Do we need at least N quark lines left?
    int nQLmin;
    
    // Particles and antenna numbers to perform the sum
    // X[0] = ISR, X[1] = FSR
    // For every set of particles there is  - one antenna function
    //                                      - the scale of the branching / clustering
    //                                      - the invariants in the order s1j, sj2, s12, S12
    //                                        (or sj2, s1j for non emission with swap)
    vector< vector< vector<Particle> > > clusteredPartsSav;
    vector< vector<        int       > > iAntSav;
    vector< vector<        double    > > scale2Sav;
    vector< vector< vector<double  > > > invariantsSav;

    // Verbose level
    int verbose;

  };

}

#endif
