// Vincia.h is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P.Z. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// This file contains global header information for Vincia.

#ifndef VINCIA_H
#define VINCIA_H

// Include Pythia 8 headers 
#include "Pythia8/Pythia.h"

// Global VINCIA functionality
#include "Vincia/VinciaCommon.h"

// Now include headers that need this information
#include "Vincia/Reconnections.h"
#include "Vincia/SubtractionTerms.h"
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"
#include "Vincia/MatchingGKS.h"
#include "Vincia/MG4interface.h"

// Define namespace inside which Vincia lives
namespace Vincia { 
  
  //*********
  
  // The VinciaPlugin class
  
  class VinciaPlugin {
    
  public:
    
    VinciaPlugin() {
      cerr << "\n (VinciaPlugin():) cannot construct VinciaPlugin :(\n\n"
      <<"  Sorry, the Vincia plugin must be constructed with a pointer"
      <<" to an\n  explicit Pythia object (since it's a plugin)." 
      <<" Please use one of \n\n"
      <<"     VinciaPlugin vincia(Pythia* pythia);\n"
      <<"     VinciaPlugin vincia(Pythia* pythia, string commandFile);\n"
      <<"     VinciaPlugin vincia(Pythia* pythia, string commandFile, string xmlDir);\n"
      <<"     VinciaPlugin vincia(Pythia* pythia, string commandFile, string xmlDir, string xmlIndexFile);\n"
      <<"\n";
    }
    
    // Constructor
    VinciaPlugin(Pythia* pythia, string commandFile="none", 
                 string xmlDir="../share/Vincia/xmldoc", 
                 string xmlIndexFile="VinciaIndex.xml") {      
      
      // First set pointers to Pythia classes
      pythiaPtr         = pythia;
      couplingsPtr      = &pythia->couplings;
      infoPtr           = &pythia->info;
      particleDataPtr   = &pythia->particleData;
      partonSystemsPtr  = &pythia->partonSystems;
      rndmPtr           = &pythia->rndm;
      settingsPtr       = &pythia->settings;
      // Initial value for pointer to user hooks.
      userHooksPtr      = 0;
      
      // Adding normalization factors here to not have to change
      // the code
      settingsPtr->addParm("Vincia:pTnormalization",4.0,true,true,4.0,4.0);   
      settingsPtr->addParm("Vincia:mDnormalization",2.0,true,true,2.0,2.0);
      settingsPtr->addParm("Vincia:pTnormalizationII",1.0,true,true,1.0,1.0);
      settingsPtr->addParm("Vincia:pTnormalizationIF",1.0,true,true,1.0,1.0);
      settingsPtr->addParm("Vincia:mSnormalizationI",1.0,true,true,1.0,1.0);
      settingsPtr->addParm("Vincia:mCnormalizationI",1.0,true,true,1.0,1.0);
      settingsPtr->addParm("Vincia:mSnormalizationF",1.0,true,true,1.0,1.0);
      
      // Find path to PYTHIA data files. 
      const char* PYTHIA8DATA = "PYTHIA8DATA";
      char* envPath = getenv(PYTHIA8DATA);
      if (envPath == 0 || *envPath == '\0') {
        cout<<" VINCIA Error: you have not set the PYTHIA8DATA environment variable.\n";
        cout<<" VINCIA Hint: if your local PYTHIA install is in the directory $HOME/yourpythia, try setting PYTHIA8DATA to $HOME/yourpythia/share/Pythia8/xmldoc"<<endl;
        cout<<"              alternatively, for a default global PYTHIA install, use /usr/share/Pythia8/xmldoc"<<endl; 
        return;
      }
      
      // Find path to VINCIA data files, i.e. xmldoc directory location.
      // Environment variable takes precedence, else use constructor input.
      string xmlPath        = "";
      bool existsVINCIADATA = false;
      const char* VINCIADATA = "VINCIADATA";
      envPath = getenv(VINCIADATA);
      if (envPath != 0 && *envPath != '\0') {
        int i = 0;
        while (*(envPath+i) != '\0') xmlPath += *(envPath+(i++));
        existsVINCIADATA = true;
      }
      else xmlPath = xmlDir;
      if (xmlPath[ xmlPath.length() - 1 ] != '/') xmlPath += "/";      
      // Set the VINCIADATA environment variable, for use in main programs
      setenv(VINCIADATA,xmlPath.c_str(),0);
      // Save the presumed location of VINCIA runtime data files
      VINCIA_SHARE = xmlPath+"../";
      // Add the VINCIA list of parameters to the PYTHIA 8 settings database
      settingsPtr->init(xmlPath+xmlIndexFile, true); 
      
      // Init FSR shower pointers and default settings
      fsrShower.functionsPtr           = &functions;
      fsrShower.resolutionPtr          = &resolution;
      fsrShower.vinCluPtr              = &vinClu;
      fsrShower.vinComPtr              = &vinCom;
      fsrShower.reconnectionsPtr       = &reconnections;
      fsrShower.ISRshowerPtr           = &isrShower;
      fsrShower.matchingPtr            = &matching;
      fsrShower.MEcorrectionsPtr       = &MECs;
      fsrShower.vetoHooksPtr           = &vetoHooks;
      fsrShower.weightsPtr             = &vinWeights;
      fsrShower.initPtr(infoPtr, settingsPtr, particleDataPtr, rndmPtr,
                        couplingsPtr, partonSystemsPtr, userHooksPtr);
      
      // Init ISR shower pointers and default settings
      isrShower.functionsPtr           = &functions;
      isrShower.resolutionPtr          = &resolution;
      isrShower.vinCluPtr              = &vinClu;
      isrShower.vinComPtr              = &vinCom;
      isrShower.reconnectionsPtr       = &reconnections;
      isrShower.FSRshowerPtr           = &fsrShower;
      isrShower.MEcorrectionsPtr       = &MECs;
      isrShower.vetoHooksPtr           = &vetoHooks;
      isrShower.weightsPtr             = &vinWeights;
      isrShower.initPtr(infoPtr, settingsPtr, particleDataPtr, rndmPtr, 
                        couplingsPtr, partonSystemsPtr, userHooksPtr);

      // Read command file 1st time
      if (commandFile != "none") pythia->readFile(commandFile); 
      verbose = 0;
      int verboseGlo = settingsPtr->mode("Vincia:verbose");
      
      if (verboseGlo >= 6)
        cout<<" VinciaPlugin() begin ------------------"<<endl;      
      
      // Vincia on/off
      vinciaOn    = settingsPtr->flag("Vincia");
      
      // Initialize standard Antennae       
      if (vinciaOn) {
        
        if (verboseGlo >= 2) {
          if (existsVINCIADATA) 
            cout<<" VinciaPlugin(): VINCIADATA environment variable ok."<<endl;
          else cout<<" VinciaPlugin(): VINCIADATA environment variable undefined. Using constructor instead."<<endl; 
        }
        if (verboseGlo >= 1) {
          cout<<" VinciaPlugin(): using configuration directory : "
          <<xmlPath<<endl;
        }
        
        if (verboseGlo >= 2)
          cout<<" VinciaPlugin(): initializing antennae..."<<endl;

        // Initialize default antenna set
        if (verboseGlo >= 3)
          cout<<" VinciaPlugin(): initializing DEF antennaSet"<<endl;
        // FSR global
        antennaSetFSRDef.initPtr(&functions, particleDataPtr, 
                                 rndmPtr, settingsPtr);
        antennaSetFSRDef.init(false,0);
        // FSR sector
        antennaSetFSRSctDef.initPtr(&functions, particleDataPtr, 
                                    rndmPtr, settingsPtr); 
        antennaSetFSRSctDef.init(true,0);
        // ISR global
        antennaSetISRDef.initPtr(&functions,  particleDataPtr, 
                                 rndmPtr, settingsPtr);
        antennaSetISRDef.init(false,0);
        
        // Initialize max antenna set
        if (verboseGlo >= 3)
          cout<<" VinciaPlugin(): initializing MAX antennaSet"<<endl;
        // FSR global
        antennaSetFSRMax.initPtr(&functions, particleDataPtr, 
                                 rndmPtr, settingsPtr);
        antennaSetFSRMax.init(false,1);
        // FSR sector
        antennaSetFSRSctMax.initPtr(&functions, particleDataPtr, 
                                    rndmPtr, settingsPtr); 
        antennaSetFSRSctMax.init(true,1);
        // ISR global
        antennaSetISRMax.initPtr(&functions, particleDataPtr, 
                                 rndmPtr, settingsPtr);
        antennaSetISRMax.init(false,1);
        
        // Initialize min antenna set
        if (verboseGlo >= 3) 
          cout<<" VinciaPlugin(): initializing MIN antennaSet"<<endl;
        // FSR global
        antennaSetFSRMin.initPtr(&functions, particleDataPtr, 
                                 rndmPtr, settingsPtr);
        antennaSetFSRMin.init(false,-1);
        // FSR sector
        antennaSetFSRSctMin.initPtr(&functions, particleDataPtr, 
                                    rndmPtr, settingsPtr); 
        antennaSetFSRSctMin.init(true,-1);
        // ISR global
        antennaSetISRMin.initPtr(&functions, particleDataPtr, 
                                 rndmPtr, settingsPtr);
        antennaSetISRMin.init(false,-1);
        
        // If filename given for tune presets, read file.
        // Get rid of this, replace with tune class instead that 
        // sets parameters and allows user mods on top.
        string tuneFile=settingsPtr->word("Vincia:tuneFile");
        if (tuneFile != "none" && tuneFile != "") { 
          // Check if this file exists, otherwise look in VINCIA_SHARE
          if (!fileExists(tuneFile)) {
            string tuneFileTmp = VINCIA_SHARE + "/tunes/" + tuneFile;
            if (fileExists(tuneFileTmp))
              tuneFile = tuneFileTmp;
          }          
          if (verboseGlo >= 3) 
            cout << " VinciaPlugin(): reading TUNE presets from "
                 <<tuneFile<<endl;
          pythia->readFile(tuneFile);
        }
        
        // Read command file 2nd time (possibility to overwrite tune settings)
        if (commandFile != "none") pythia->readFile(commandFile); 
        
        // Set pointers for global and sector antenna sets
        // Initialization done when shower initialize
        // FSR
        antennaSetFSR.initPtr(&functions, particleDataPtr, 
                              rndmPtr, settingsPtr);
        antennaSetFSRSct.initPtr(&functions, particleDataPtr, 
                                 rndmPtr, settingsPtr);
        // ISR
        antennaSetISR.initPtr(&functions, particleDataPtr, 
                              rndmPtr, settingsPtr);
        
        // Now set verbose
        verbose = verboseGlo;
        
        if (verboseGlo >= 2)
          cout<<" VinciaPlugin(): initializing pointers..."<<endl;
        
        // Hand antenna set pointers to shower and matching objects
        // FSR
        fsrShower.initAntPtr(&antennaSetFSR,&antennaSetFSRDef,
                             &antennaSetFSRMax,&antennaSetFSRMin);
        matching.initAntPtr(&antennaSetFSR,&antennaSetFSRDef,
                            &antennaSetFSRMax,&antennaSetFSRMin);
        fsrShower.initAntSctPtr(&antennaSetFSRSct,&antennaSetFSRSctDef,
                                &antennaSetFSRSctMax,&antennaSetFSRSctMin);
        matching.initAntSctPtr(&antennaSetFSRSct,&antennaSetFSRSctDef,
                               &antennaSetFSRSctMax,&antennaSetFSRSctMin);
        // ISR
        isrShower.initAntPtr(&antennaSetISR,&antennaSetISRDef,
                             &antennaSetISRMax,&antennaSetISRMin);
        MECs.initAntPtr(&antennaSetFSR,&antennaSetISR,
                        &antennaSetFSRDef,&antennaSetISRDef,
                        &antennaSetFSRMax,&antennaSetISRMax,
                        &antennaSetFSRMin,&antennaSetISRMin);
        
        
        // Set Pythia 8 shower pointers to Vincia
        pythia->setShowerPtr(&fsrShower, &fsrShower, &isrShower);
        
        // Set the veto hooks
        pythia->setUserHooksPtr( &vetoHooks );
      } 
      
      // If VINCIA off, let us know about it
      else {
        if (verbose >= 2) printOut("VinciaPlugin","Vincia is switched off");
      }
      
      // Pass pointers on to objects that require them
      resolution.initPtr(settingsPtr);
      rambo.initPtr(rndmPtr);
      subtractionTerms.initPtr(settingsPtr);
      vinCom.initPtr(particleDataPtr,settingsPtr);
      vinClu.initPtr(rndmPtr,settingsPtr,&vinCom);
      mgInterface.initPtr(couplingsPtr,particleDataPtr,settingsPtr,&vinCom);
      matching.initPtr(&mgInterface, particleDataPtr, &resolution,
                       rndmPtr, settingsPtr, 
                       &subtractionTerms, &vinClu, &vinCom, &functions);    
      reconnections.initPtr(rndmPtr,settingsPtr);
      MECs.initPtr(settingsPtr, &mgInterface, rndmPtr, &resolution, &vinCom, &vinClu);
      vinWeights.initPtr(settingsPtr);
      
      if (verboseGlo >= 6) 
        cout<<" VinciaPlugin() end ------------------"<<endl;      
    }
    
    // Initialize a few objects, extra function to allow use of
    // pythia.readString in main program
    void init() {
      vinCom.init();
      resolution.init();
      reconnections.init();
      MECs.init();
      vinClu.init();
      subtractionTerms.init();
      vinWeights.init();
      pythiaPtr->init();
    }
    
    //*********
    
    // Automatically set verbose level in all inheriting members
    void setVerbose(int verboseIn) {
      vinCom.setVerbose(verboseIn);
      resolution.setVerbose(verboseIn);
      fsrShower.setVerbose(verboseIn);
      isrShower.setVerbose(verboseIn);
      matching.setVerbose(verboseIn);
      reconnections.setVerbose(verboseIn);
      mgInterface.setVerbose(verboseIn);
      vinClu.setVerbose(verboseIn);
    }
    
    // Main VINCIA on/off switch
    bool vinciaOn;
    
    // Instantiate Vincia Objects
    VinciaCommon        vinCom;
    Resolution          resolution;
    TimeAntShower       fsrShower;
    SpaceAntShower      isrShower;
    MatchingGKS         matching;
    Reconnections       reconnections;
    SubtractionTerms    subtractionTerms;
    VinciaVetoHooks     vetoHooks;
    VinciaWeights       vinWeights;
    MEcorrections       MECs;

    // Auxiliary objects
    Functions           functions;
    MG4interface        mgInterface;
    Rambo               rambo;
    VinClu              vinClu;  
    
    // Vectors of physical Antennae with uncertainty bands
    AntennaSetFSR antennaSetFSR;
    AntennaSetFSR antennaSetFSRDef;
    AntennaSetFSR antennaSetFSRMax;
    AntennaSetFSR antennaSetFSRMin;
    // Sector Antennae
    AntennaSetFSR antennaSetFSRSct;
    AntennaSetFSR antennaSetFSRSctDef;
    AntennaSetFSR antennaSetFSRSctMax;
    AntennaSetFSR antennaSetFSRSctMin;
    
    // Vectors of physical Antennae with uncertainty bands
    AntennaSetISR antennaSetISR;
    AntennaSetISR antennaSetISRDef;
    AntennaSetISR antennaSetISRMax;
    AntennaSetISR antennaSetISRMin;

    // Location of runtime VINCIA files
    string VINCIA_SHARE;
    
    // Pointers to Pythia classes
    Pythia*        pythiaPtr;
    Couplings*     couplingsPtr;
    Info*          infoPtr;
    ParticleData*  particleDataPtr;
    PartonSystems* partonSystemsPtr;
    Rndm*          rndmPtr; 
    Settings*      settingsPtr; 
    UserHooks*     userHooksPtr;
    
    // Replacing the internal vincia hooks pointer with another one, 
    // to allow additional user hooks
    // NOTE vetoHooksPtrIn needs to have the same functionality as 
    // the original one, otherwise  vetoes for event samples are wrong
    void replaceVetoHooks(Pythia* pythia, VinciaVetoHooks* vetoHooksPtrIn) {
      fsrShower.vetoHooksPtr = vetoHooksPtrIn;
      isrShower.vetoHooksPtr = vetoHooksPtrIn;
      pythia->setUserHooksPtr(vetoHooksPtrIn);
    }
    
    // Shortcut for post-Vincia event weight
    double weight(int iVar=0) {
      if (settingsPtr->flag("Vincia"))
        return vinWeights.getWeight(iVar);
      return infoPtr->weight();
    }
    double weightInit() {
      if (settingsPtr->flag("Vincia"))
        return vinWeights.getWeightInit();
      return infoPtr->weight();
    }
    int nVariations() { return vinWeights.nVariations(); }
    bool weightingOccurred()   { return vinWeights.weightingOccurred();   }
    bool reweightingOccurred() { return vinWeights.reweightingOccurred(); }
    
    // Utilities for printing Info and Internal histograms
    void printInfo( ostream& os = cout ) { 
      fsrShower.printInfo(os, true);
      isrShower.printInfo(os, true);
    }
    void printHistos( ostream& os = cout ) { fsrShower.printHistos(os); }
    void writeHistos( string fileName = "vincia", string lastName = "dat" ) { 
      fsrShower.writeHistos(fileName,lastName); 
    }
    const Hist& getDiagnosticHistogram( string name ) {
      return fsrShower.getDiagnosticHistogram(name);
    }
    
    private :

    // Verbose level
    int verbose;
    
  }; // end Vincia plugin
  
} // end Vincia namespace
#endif
