// VRfile.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// Copyright (C) 2016 P. Skands.
// VINCIA is licenced under the GNU GPL version 2.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Function definitions (not found in the header) for the VRfile class 

// Include the VRfile header
#include "VinciaRoot/VRfile.h"

namespace Vincia {

  // Read data points from file
  bool VRfile::readFile(string fileNameIn, int verboseIn) {
  
    // Initialize
    init(verboseIn);
    
    // Verbose output
    if (verbose >= 3) cout<<"(VRfile): begin read "<<fileNameIn<<endl;
    
    // Open input stream. Return false if fails
    ifstream inStream;
    inStream.open(fileNameIn.c_str());
    if (! inStream) return false;
    
    // Read lines, return false if no useful lines
    int nLines = 0;
    while (1) {
      if ( readLine(inStream)) ++nLines; 
      if ( !inStream.good() ) break;
    }
    if (nLines <= 0) {
      if (verbose >= 1) 
        cout<<"(VRfile): warning! no readable lines in "<<fileNameIn<<endl;
      return false;
    }
    // Verbose output
    if (verbose >= 2) {
      cout<<"(VRfile): "<<nLines<<" readable data points in "
          <<fileNameIn<<endl;
    }
    
    // Read SRC, REF, and ECM if present
    inStream.clear();
    char line[256];
    bool foundSRC = false;
    bool foundREF = false;
    while (1) {
      inStream.getline(line,256);
      if (inStream.good()) {
        string sline(line);                    
        int ipad = 1;
        if (sline.substr(sline.find("=")+1,1) == " ") {
          ipad = 2;
          if (sline.substr(sline.find("=")+2,1) == " ") {
            ipad = 3;
            if (sline.substr(sline.find("=")+3,1) == " ") {
              ipad = 4;
              if (sline.substr(sline.find("=")+4,1) == " ") {
                ipad = 5;
              }
            }
          }
        }
        // Use first occurrence of SRC and REF
        if (sline.find("SRC") != string::npos && !foundSRC) {
          fileSRC  = sline.substr(sline.find("=")+ipad,sline.size());
          foundSRC = true;
        } else if (sline.find("REF") != string::npos && !foundREF) {
          fileREF = sline.substr(sline.find("=")+ipad,sline.size());
          foundREF = true;
        } else if (sline.find("SQRT(S)") != string::npos ||
                   sline.find("ECM") != string::npos ) {
          fileECM = sline.substr(sline.find(":")+ipad,sline.size());
        }
      } else {
        break;
      }
    }
    
    // Finalize read-in (compute sum of weights, etc)
    bool ok = finalize();

    // Close file
    inStream.close();

    // Everything ok?
    return ok;
  }

  // Read line from data file. 
  bool VRfile::readLine(ifstream& inStream) {

    // Sanity check
    if (format.size() == 0) return false;
    
    // Read values into buffer (default: assume unit-wide bins)
    double val, xL(0.5), xC, xR(0.5), y(0.);
    double sigStatHi(0.), sigStatLo(0.), sigSysHi(0.), sigSysLo(0.);
    double sigTotHi(0.), sigTotLo(0.), sigSys2Hi(0.), sigSys2Lo(0.);
    double scalFac(1.);
    string genericString;
    bool hasX          = false;
    bool hasDX         = true;
    bool hasStatErr    = false;
    bool hasSysErr     = false;
    bool hasSysErr2    = false;
    bool hasTotErr     = false;   
    bool logx          = false;
    bool linearSysErr  = false;
    bool linearSysErr2 = false;
    for (unsigned int iFmt=0; iFmt<format.size(); ++iFmt) { 
      // Format specifiers 1-9 are for strings:
      if (format[iFmt] < 10 && format[iFmt] >= 0) {
        inStream >> genericString ;        
      } 
      // All other format specifiers are treated as double:
      else {
        inStream >> val;
      }
      if (!inStream.good()) {
        if (format[iFmt] == 0) continue;
        else return false;
      }
      if (format[iFmt] == 0) continue;
      if (abs(format[iFmt]) < 10) continue;
      if (format[iFmt] >= 10 && format[iFmt] <= 19) {
        if (format[iFmt] == 10 || format[iFmt] == 12 || format[iFmt] == 13 || format[iFmt] == 14) hasDX = false;
        if (format[iFmt] <= 11) xL = val;
        else if (format[iFmt] == 12) xR = 1.0 - val;
        else if (format[iFmt] == 13) xL = (val > 0.) ? log(val) : -1e9;
        else if (format[iFmt] == 14) xL = (val > 0.) ? log10(val) : -1e9;
      } 
      else if (format[iFmt] >= 20 && format[iFmt] <= 29) {
        hasX = true;
        xC = val;
        if (format[iFmt] == 22) xC = 1.0 - val;
        else if (format[iFmt] == 23) xC = (val > 0.) ? log(val) : -1e9;
        else if (format[iFmt] == 24) xC = (val > 0.) ? log10(val) : -1e9;
        else if (format[iFmt] == 25) xC = xCsav.size();
      } 
      else if (format[iFmt] >= 30 && format[iFmt] <= 39) {
        if (format[iFmt] == 30) hasDX = false;
        if (format[iFmt] <= 31) xR = val;
        else if (format[iFmt] == 32) xL = 1.0 - val;
        else if (format[iFmt] == 33) xR = (val > 0.) ? log(val) : -1e9;
        else if (format[iFmt] == 34) xR = (val > 0.) ? log10(val) : -1e9;
      }
      else if (format[iFmt] >= 40 && format[iFmt] <= 49) {
        if (format[iFmt] == 45) scalFac = val;
        else if (format[iFmt] == 46 && val != 0.0) scalFac = 1./val;
        else {
          y  = val;
	  if (format[iFmt] == 43) logx = true;
        }
      }
      // Stat uncertainties:
      // 50-52 are interpreted as absolute
      // 55-57 are interpreted as relative (fraction of y value)
      // 552,562,572 are interpreted as per cent
      // 553,563,573 are interpreted as per mille
      else if (format[iFmt]%100 >= 50 && format[iFmt]%100 <= 59) {
        hasStatErr = true;
        if (format[iFmt] == 50 || format[iFmt] == 51) sigStatHi = val;
        if (format[iFmt] == 50 || format[iFmt] == 52) sigStatLo = val;
        if (format[iFmt] == 55 || format[iFmt] == 56) sigStatHi = val*y;
        if (format[iFmt] == 55 || format[iFmt] == 57) sigStatLo = val*y;
        if (format[iFmt] == 552 || format[iFmt] == 562)
          sigStatHi = val*y*0.01;
        if (format[iFmt] == 552 || format[iFmt] == 572)
          sigStatLo = val*y*0.01;
        if (format[iFmt] == 553 || format[iFmt] == 563)
          sigStatHi = val*y*0.001;
        if (format[iFmt] == 553 || format[iFmt] == 573)
          sigStatLo = val*y*0.001;
      }
      // Systematic errors: > 0 (<0) indicates sum in quadrature (linearly)
      // 60-62 are interpreted as absolute
      // 65-67 are interpreted as relative (fraction of y value)
      else if (abs(format[iFmt]) >= 60 && abs(format[iFmt]) <= 69) {
        hasSysErr  = true;
        if (abs(format[iFmt]) == 60 || abs(format[iFmt]) == 61)
          sigSysHi = val;
        if (abs(format[iFmt]) == 60 || abs(format[iFmt]) == 62)
          sigSysLo = val;
        if (abs(format[iFmt]) == 65 || abs(format[iFmt]) == 66)
          sigSysHi = y*val;
        if (abs(format[iFmt]) == 65 || abs(format[iFmt]) == 67)
          sigSysLo = y*val;
        if (abs(format[iFmt]) == 652 || abs(format[iFmt]) == 662)
          sigSysHi = y*val*0.01;
        if (abs(format[iFmt]) == 652 || abs(format[iFmt]) == 672)
          sigSysLo = y*val*0.01;
        if (abs(format[iFmt]) == 653 || abs(format[iFmt]) == 663)
          sigSysHi = y*val*0.001;
        if (abs(format[iFmt]) == 653 || abs(format[iFmt]) == 673)
          sigSysLo = y*val*0.001;
        if (format[iFmt] < 0) linearSysErr = true;
      }
      else if (abs(format[iFmt]) >= 70 && abs(format[iFmt]) <= 79) {
        hasSysErr2 = true;
        if (abs(format[iFmt]) == 70 || abs(format[iFmt]) == 71)
          sigSys2Hi = val;
        if (abs(format[iFmt]) == 70 || abs(format[iFmt]) == 72)
          sigSys2Lo = val;
        if (abs(format[iFmt]) == 75 || abs(format[iFmt]) == 76)
          sigSys2Hi = y*val;
        if (abs(format[iFmt]) == 75 || abs(format[iFmt]) == 77)
          sigSys2Lo = y*val;
        if (abs(format[iFmt]) == 752 || abs(format[iFmt]) == 762)
          sigSys2Hi = y*val*0.01;
        if (abs(format[iFmt]) == 752 || abs(format[iFmt]) == 772)
          sigSys2Lo = y*val*0.01;
        if (abs(format[iFmt]) == 753 || abs(format[iFmt]) == 763)
          sigSys2Hi = y*val*0.001;
        if (abs(format[iFmt]) == 753 || abs(format[iFmt]) == 773)
          sigSys2Lo = y*val*0.001;
        if (format[iFmt] < 0) linearSysErr2 = true;
      }
      else if (format[iFmt] >= 80 && format[iFmt] <= 89) {
        hasTotErr  = true;
        if (abs(format[iFmt]) == 80 || abs(format[iFmt]) == 81)
          sigTotHi = val;
        if (abs(format[iFmt]) == 80 || abs(format[iFmt]) == 82)
          sigTotLo = val;
        if (abs(format[iFmt]) == 85 || abs(format[iFmt]) == 86)
          sigTotHi = y*val;
        if (abs(format[iFmt]) == 85 || abs(format[iFmt]) == 87)
          sigTotLo = y*val;
        if (abs(format[iFmt]) == 852 || abs(format[iFmt]) == 862)
          sigTotHi = y*val*0.01;
        if (abs(format[iFmt]) == 852 || abs(format[iFmt]) == 872)
          sigTotLo = y*val*0.01;
        if (abs(format[iFmt]) == 853 || abs(format[iFmt]) == 863)
          sigTotHi = y*val*0.001;
        if (abs(format[iFmt]) == 853 || abs(format[iFmt]) == 873)
          sigTotLo = y*val*0.001;
        if (abs(format[iFmt]) < 0) linearSysErr = true;
      }
    }
    // Skip any additional content until end of line
    char next;
    while(inStream.get(next))
      {
        if (next == '\n')  // If the file has been opened in
          {    break;        // text mode then it will correctly decode the
          }                  // platform specific EOL marker into '\n'
      }  
  
     // If dxL and dxR given, construct xL and xR from xC
    if (hasDX) {
      if (!hasX) cout<<"(VRfile): cannot construct bin edges"
                   " without central X value"<<endl;
      xL = xC - abs(xL);
      xR = xC + abs(xR);
    }
    // If only xL and xR given, construct xC
    if (!hasX) {
      xC = (xL + xR)/2.;
    }
    // If only 2nd sys error given, move to 1st
    if (hasSysErr2 && !hasSysErr) {
      hasSysErr2    = false;
      hasSysErr     = true;
      linearSysErr  = linearSysErr2;
      linearSysErr2 = false;
      sigSysLo   = sigSys2Lo;
      sigSysHi   = sigSys2Hi;
      sigSys2Lo  = 0.;
      sigSys2Hi  = 0.;
    }
    // Construct total error, summing stat and sys errors as requested
    if (!hasTotErr) {
      double sigSysTotLo = abs(sigSysLo);
      double sigSysTotHi = abs(sigSysHi);
      // Add individual systematic components linearly or qudratically
      if (hasSysErr2) {
        if (linearSysErr2) { 
          sigSysTotLo += abs(sigSys2Lo);
          sigSysTotHi += abs(sigSys2Hi);
        } else {
          sigSysTotLo = sqrt(pow(sigSysLo,2.) + pow(sigSys2Lo,2.));
          sigSysTotHi = sqrt(pow(sigSysHi,2.) + pow(sigSys2Hi,2.));
        }
      }
      // Add syst errors linearly or quadratically to stat error
      if (linearSysErr) {
        sigTotLo = abs(sigStatLo) + abs(sigSysTotLo);
        sigTotHi = abs(sigStatHi) + abs(sigSysTotHi);
      } else {
        sigTotLo = sqrt(pow(sigStatLo,2.)+pow(sigSysTotLo,2.));
        sigTotHi = sqrt(pow(sigStatHi,2.)+pow(sigSysTotHi,2.));
      }
    }
    // If a total error error is given, with stat component, assume the rest
    // is systematic.
    else if (hasStatErr) {      
      if (linearSysErr) {
        sigSysLo = max(0.0,abs(sigTotLo) - abs(sigStatLo));
        sigSysHi = max(0.0,abs(sigTotHi) - abs(sigStatHi));
      } else {
        sigSysLo = sqrt(max(0.0,pow(sigTotLo,2.) - pow(sigStatLo,2.)));
        sigSysHi = sqrt(max(0.0,pow(sigTotHi,2.) - pow(sigStatHi,2.)));
      }
    }
    // If total error is given, with systematic component, assume the rest
    // is statistical 
    else if (hasSysErr) {
      double sigSysSumLo = abs(sigSysLo);
      double sigSysSumHi = abs(sigSysHi);
      if (hasSysErr2) {
        if (linearSysErr2) {
          sigSysSumLo += abs(sigSys2Lo);
          sigSysSumHi += abs(sigSys2Hi);
        } else {
          sigSysSumLo = sqrt(pow(sigSysLo,2.) + pow(sigSys2Lo,2.));
          sigSysSumHi = sqrt(pow(sigSysHi,2.) + pow(sigSys2Hi,2.));
        }
      }
      if (linearSysErr) {
        sigStatLo = max(0.0,abs(sigTotLo)-abs(sigSysSumLo));
        sigStatHi = max(0.0,abs(sigTotHi)-abs(sigSysSumHi));
      } else {
        sigStatLo = sqrt(max(0.0,pow(sigTotLo,2.)-pow(sigSysSumLo,2.)));
        sigStatHi = sqrt(max(0.0,pow(sigTotHi,2.)-pow(sigSysSumHi,2.)));
      }
    }
    // If only total error is given, assign all of it as "systematic"
    // (to avoid a purely "statistical" interpretation)
    else {
      sigSysLo = sigTotLo;
      sigSysHi = sigTotHi;
    }
    
    // Optionally suppress zero values with zero uncertainty
    // if ( y == 0.0 && sigTotLo == 0.0 && sigTotHi == 0.0 ) return true;
    
    // Check if y values to be scaled by x values (log x axis scaling)
    if (logx) {
      y /= exp(-xC);
      sigTotLo  /= exp(-xC);
      sigTotHi  /= exp(-xC);
      sigStatLo /= exp(-xC);
      sigStatHi /= exp(-xC);
      sigSysLo  /= exp(-xC);
      sigSysHi  /= exp(-xC);
      sigSys2Lo /= exp(-xC);
      sigSys2Hi /= exp(-xC);
    }

    // Check if applying a scale factor
    if (scalFac != 1.0) {
      y *= scalFac;
      sigTotLo  *= scalFac;
      sigTotHi  *= scalFac;
      sigStatLo *= scalFac;
      sigStatHi *= scalFac;
      sigSysLo  *= scalFac;
      sigSysHi  *= scalFac;
      sigSys2Lo *= scalFac;
      sigSys2Hi *= scalFac;
    }

    // Add data point
    // assume x values are ordered ascending or descending, find out which
    if (xCsav.empty() || xC < *(xCsav.begin())) {
      // insert at the beginning
      xLsav.insert(xLsav.begin(), xL);
      xCsav.insert(xCsav.begin(), xC);
      xRsav.insert(xRsav.begin(), xR);
      ySav.insert(ySav.begin(), y);
      uTotLoSav.insert(uTotLoSav.begin(), sigTotLo);
      uTotHiSav.insert(uTotHiSav.begin(), sigTotHi);
      uStatLoSav.insert(uStatLoSav.begin(), abs(sigStatLo));
      uStatHiSav.insert(uStatHiSav.begin(), abs(sigStatHi));
      uSysLoSav.insert(uSysLoSav.begin(), abs(sigSysLo));
      uSysHiSav.insert(uSysHiSav.begin(), abs(sigSysHi));
      uSys2LoSav.insert(uSys2LoSav.begin(), abs(sigSys2Lo));
      uSys2HiSav.insert(uSys2HiSav.begin(), abs(sigSys2Hi));
    } else {
      // insert at the end
      xLsav.push_back(xL);
      xCsav.push_back(xC);
      xRsav.push_back(xR);
      ySav.push_back(y);
      uTotLoSav.push_back(sigTotLo);
      uTotHiSav.push_back(sigTotHi);
      uStatLoSav.push_back(abs(sigStatLo));
      uStatHiSav.push_back(abs(sigStatHi));
      uSysLoSav.push_back(abs(sigSysLo));
      uSysHiSav.push_back(abs(sigSysHi));
      uSys2LoSav.push_back(abs(sigSys2Lo));
      uSys2HiSav.push_back(abs(sigSys2Hi));
    }
    
    
    // Store number of systematic uncertainties
    nSysSav = 0;
    if (hasSysErr) ++nSysSav;
    if (hasSysErr2) ++nSysSav;
    
    // Line read ok
    return true;
  }
  
  // Finalize read-in: fill in empty arrays and compute normalization.
  // Assumes x values are ascending
  bool VRfile::finalize() {
    
    // Zero weight sum
    sumWeights = 0.;
    
    // Loop over data points        
    for (int i=0; i<int(ySav.size()); ++i) {
      // Check for overlapping bins and restrict if necessary
      if (i+1 < int(ySav.size())) {
        if (xRsav[i] > xLsav[i+1]) {
          xRsav[i]   = 0.5 * (xCsav[i] + xCsav[i+1]);
          xLsav[i+1] = xRsav[i];
          if (i==0) xLsav[i]   = xCsav[i] - (xRsav[i] - xCsav[i]);
          if (i+1==int(ySav.size())-1) 
            xRsav[i+1] = xCsav[i+1] + (xCsav[i+1] - xLsav[i+1]);
        }
      }
      double dx = (xRsav[i]-xLsav[i]);
      sumWeights += ySav[i] * dx;
    }
    
    // Everything ok.
    return true;
  }
  
} // end namespace VINCIA
