// AntennaSets.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Include the AntennaSets header file
#include "Vincia/AntennaSets.h"

namespace Vincia {
  
  //**********************************************************************
  //==========================================================================
  // AntennaSetFSR Methods

  void AntennaSetFSR::init(bool isSectorIn, int variation) {

    if (! isInitPtr) {
      printOut("AntennaSetFSR::init","Cannot initialize, pointers not set.");
      return;
    }
    
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // Create antenna objects
    antPtrs.resize(5);
    antPtrs[iQQemit]  = new QQEmit();
    antPtrs[iQGemit]  = new QGEmit();
    antPtrs[iGGemit]  = new GGEmit();
    antPtrs[iQGsplit] = new QGSplit();
    antPtrs[iGGsplit] = new GGSplit();
    
    // Loop through antPtrs and initialise
    for (vector<AntennaFSR*>::iterator antPtrIt = antPtrs.begin(); 
         antPtrIt != antPtrs.end(); ++antPtrIt) {

      // Initialize antenna pointers to PYTHIA 8 objects
      (*antPtrIt)->initPtr(functionsPtr, particleDataPtr, rndmPtr, settingsPtr);
      
      // Initialize antenna itself
      bool pass = (*antPtrIt)->init(isSectorIn,variation);
      
      // Optionally check antenna (singularities, positivity, etc)
      if (settingsPtr->flag("Vincia:checkAntennae")) 
        pass = pass && (*antPtrIt)->check();
      
      // Everything OK?
      if (pass) {
        if (verbose >= 2)
          printOut("Vincia::AntennaSetFSR::init","Added to antenna list: "
                   +(*antPtrIt)->humanName());
      } 
      else if (verbose >= 1) {
        printErr("Vincia::AntennaSetFSR::init","Added to antenna list: "
                 +(*antPtrIt)->humanName()+
                 " (but one or more consistency checks failed)");
      }
    }
  }

  //**********************************************************************
  // AntennaSetISR Methods

  void AntennaSetISR::init(bool isSectorIn, int variation) {

    if (! isInitPtr) {
      printOut("AntennaSetISR::init","Cannot initialize, pointers not set.");
      return;
    }
    
    // Set verbose level
    verbose = settingsPtr->mode("Vincia:verbose");

    // Create antenna objects
    antPtrs.resize(12);
    antPtrs[iQQemitII]  = new QQEmitII();
    antPtrs[iGQemitII]  = new GQEmitII();
    antPtrs[iGGemitII]  = new GGEmitII();
    antPtrs[iQXsplitII] = new QXSplitII();
    antPtrs[iGXconvII]  = new GXConvII();
    antPtrs[iQQemitIF]  = new QQEmitIF();
    antPtrs[iQGemitIF]  = new QGEmitIF();
    antPtrs[iGQemitIF]  = new GQEmitIF();
    antPtrs[iGGemitIF]  = new GGEmitIF();
    antPtrs[iQXsplitIF] = new QXSplitIF();
    antPtrs[iGXconvIF]  = new GXConvIF();
    antPtrs[iXGsplitIF] = new XGSplitIF();
    
    // Loop through antPtrs and initialise
    for (vector<AntennaISR*>::iterator antPtrIt = antPtrs.begin(); 
         antPtrIt != antPtrs.end(); ++antPtrIt) {

      // Initialize antenna pointers to PYTHIA 8 objects
      (*antPtrIt)->initPtr(functionsPtr, particleDataPtr, rndmPtr, settingsPtr);
      
      // Initialize antenna itself
      bool pass = (*antPtrIt)->init(isSectorIn,variation);
      
      // Optionally check antenna (singularities, positivity, etc)
      if (settingsPtr->flag("Vincia:checkAntennae")) 
        pass = pass && (*antPtrIt)->check();
      
      // Everything OK?
      if (pass) {
        if (verbose >= 2)
          printOut("Vincia::AntennaSetISR::init","Added to antenna list: "
                   +(*antPtrIt)->vinciaName());
      } 
      else if (verbose >= 1) {
        printErr("Vincia::AntennaSetISR::init","Added to antenna list: "
                 +(*antPtrIt)->vinciaName()+
                 " (but one or more consistency checks failed");
      }
    }
  }
  
} // End namespace Vincia
