// Merging.cc is a part of the VINCIA plug-in to the PYTHIA 8 generator.
// VINCIA is licenced under the GNU GPL version 2.
// Copyright (C) 2016 P. Skands.
// Please respect the MCnet Guidelines for Event Generator Authors and Users, 
//               http://www.montecarlonet.org/GUIDELINES

// Merging function definitions for the TimeAntShower and SpaceAntShower 
// classes, added by NF.
// NOTE Only for Vincia:orderingMode = 1
//               Vincia:hardJets = off
//               Vincia:hardJetsQCD = off
//    TBD        Vincia:GluonSplittingCorrection = off
//    TBD        Vincia:hyperjet = off (default)
//    TBD        Vincia:massCorrectionSplitF < 2 (default)
//    TBD        Vincia:massCorrectionEmitF = 0 (default)
// TODO since merging uses pythia's alphaS ptr the settings need to be
// same, also CMW needs to be off!!! maybe it actually works with
// pythias off, ours on
// NOTE Assuming FSR pTnormalization = 4, mDnormalization = 2, all other = 1,
//      as defaults set in Vincia.h
// TODO Is the factor 1/2 in gluon emission correct?
// TODO why does Z sometimes disappear from event record?

// Include the Vincia headers 
#include "Vincia/TimeAntShower.h"
#include "Vincia/SpaceAntShower.h"

namespace Vincia {
  
  const bool PRINTNOW         = false;
  const bool PRINTDETAILSNOW  = false;
  
  //**********************************************************************
  // TimeAntShower
  //**********************************************************************
  
  // Return clustering kinematics - as needed for merging.
  Event TimeAntShower::clustered( const Event& eventOld, int iRa, 
                                  int iEm, int iRe, string name ) {
    if (!isTimelike(eventOld, iRa, iEm, iRe, name)) return Event();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return Event();
    
    Event event = eventOld;
    
    Particle radiator = event[iRa];
    Particle emission = event[iEm];
    Particle recoiler = event[iRe];
    
    // Clustering: kinematic part
    vector<int> idsVc;
    vector<Vec4> pVc;
    // Gluon emission or quark from gluon splitting (ij)k
    if (emission.id() == 21 || emission.id() > 0) {
      idsVc.push_back(radiator.id());
      idsVc.push_back(emission.id());
      idsVc.push_back(recoiler.id());
      pVc.push_back(radiator.p());
      pVc.push_back(emission.p());
      pVc.push_back(recoiler.p());
    }
    // Antiquark from gluon splitting i(jk)
    else {
      idsVc.push_back(recoiler.id());
      idsVc.push_back(emission.id());
      idsVc.push_back(radiator.id());
      pVc.push_back(recoiler.p());
      pVc.push_back(emission.p());
      pVc.push_back(radiator.p());
    }
    int ii = 0;
    int ij = 1;
    int ik = 2;
    vector<int> idClu; 
    vector<Vec4> pClu;
    int iant;
    bool swapAB;
    if (!vinCluPtr->getClu(idsVc, pVc, ii, ij, ik, idClu,
                           pClu, iant, swapAB)) return Event();
    
    // Id's and colours
    int idj   = emission.id();
    if (idj == 21) swapAB = true;
    int colj  = emission.col();
    int acolj = emission.acol();
    int idI   = idClu[swapAB ? 0 : 1]; // radiator
    int colI  = radiator.col();
    int acolI = radiator.acol();
    int idK   = idClu[swapAB ? 1 : 0]; // recoiler
    int colK  = recoiler.col();
    int acolK = recoiler.acol();
    // Gluon emission
    if (idj == 21) {
      if (colj == acolK) acolK = colI;
      else if (acolj == colK) colK = acolI;
    }
    // Gluon splitting
    else {
      idI = 21;
      if (colI == 0) colI = colj;
      if (acolI == 0) acolI = acolj;
    }
    
    // Change id, momentum, colour
    // Replace radiator and recoiler in event record
    event[iRa].id(idI);
    event[iRa].p(pClu[swapAB ? 0 : 1]);
    event[iRa].col(colI);
    event[iRa].acol(acolI);
    
    event[iRe].id(idK);
    event[iRe].p(pClu[swapAB ? 1 : 0]);
    event[iRe].col(colK);
    event[iRe].acol(acolK);
      
    // Remove emission from event record
    event.remove(iEm, iEm);
    // Remove emission from mother/daughter lists
    for (int i=0; i<event.size(); i++) {
      // Mother1
      if (event[i].mother1() > iEm) 
        event[i].mother1(event[i].mother1()-1);
      else if (event[i].mother1() == iEm) {
        if (event[i].mother2() == iEm) 
          event[i].mothers(0, 0);
        else if (event[i].mother2() > iEm)
          event[i].mothers(event[i].mother2()-1, 0);
        else
          event[i].mothers(event[i].mother2(), 0);
      }
      // Mother2
      if (event[i].mother2() > iEm) 
        event[i].mother2(event[i].mother2()-1);
      else if (event[i].mother2() == iEm) 
        event[i].mother2(0);
      // Daugther1
      if (event[i].daughter1() > iEm) 
        event[i].daughter1(event[i].daughter1()-1);
      else if (event[i].daughter1() == iEm) {
        if (event[i].daughter2() == iEm)
          event[i].daughters(0, 0);
        else if (event[i].daughter2() > iEm)
          event[i].daughters(iEm, event[i].daughter2()-1);
        else
          event[i].daughters(event[i].daughter2(), 0);
      }
      // Daugther2
      if (event[i].daughter2() > iEm) 
        event[i].daughter2(event[i].daughter2()-1);
      else if (event[i].daughter2() == iEm) {
        if (event[i].daughter1() < iEm-1)
          event[i].daughters(event[i].daughter1(), iEm-1);
        else 
          event[i].daughter2(0);
      }
    }
    
    // Check energy-momentum conservation
    if (verbose >= 3)  {
      // Start with incoming partons = beam daughters
      Vec4 pTot = -(event[event[1].daughter1()].p() + 
                    event[event[2].daughter1()].p());
      double eLab = abs(pTot.e());
      // Sum final state
      for (int i=0; i<event.size(); ++i)
        if (event[i].isFinal()) pTot += event[i].p();
      // Check energy-momentum/charge conservation.
      double epDev = abs( pTot.e()) + abs(pTot.px()) + abs(pTot.py())
                                    + abs(pTot.pz() );
      if (epDev > vinComPtr->epTolErr * eLab) {
        cout << "ERROR in TimeAntShower::clustered for " << name
             << ": energy-momentum not conserved" << endl;
//         cout << " Unclustered event " << name << ":" << endl;
//         eventOld.list();
//         cout << " Clustered event for " << name << ":" << endl;
//         event.list();
      }
      else if (epDev > vinComPtr->epTolWarn * eLab) {
        cout << "WARNING in TimeAntShower::clustered for " << name
             << ": energy-momentum not quite conserved" << endl;
      }
    }
    
    if (PRINTNOW) {
      cout << " Unclustered event for " << name << ":" << endl;
      eventOld.list();
      cout << " Clustered event for " << name << ":" << endl;
      event.list();
    }
    
    return event;
  }
  
  // Return the evolution variable.
  // Important note:
  // - This map must contain an entry for the shower evolution variable, 
  //   specified with key "t".
  // - This map must contain an entry for the shower evolution variable from
  //   which the shower would be restarted after a branching. This entry
  //   must have key "tRS", 
  // - This map must contain an entry for the argument of \alpha_s used
  //   for the branching. This entry must have key "scaleAS". 
  // - This map must contain an entry for the argument of the PDFs used
  //   for the branching. This entry must have key "scalePDF". 
  /*map<string, double> TimeAntShower::getStateVariables( const Event& event, int iRa, 
                                                        int iEm, int iRe, string name) {
    map<string, double> evolVals;
    evolVals.clear();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return evolVals;
    
    int idEm = event[iEm].idAbs();
    
    // Evolution variable
    double Q2evol  = 0.0;
    double kMu2    = 0.0;
    double normFac = 1.0;
    
    // 1. Gluon emission: symmetric
    if (idEm == 21) {
      Q2evol  = resolutionPtr->Q2E(event[iRa], event[iEm], event[iRe], 
                                   evolutionType);
      kMu2    = aSkMu2;
      normFac = ((evolutionType == 1) ? pTnormalization : mDnormalization);
    }
    
    // 2. Gluon splitting: sjk
    else if (idEm < 7) {
      Q2evol  = resolutionPtr->Q2E(event[iRe], event[iEm], event[iRa], 7);
      kMu2    = aSkMu2Split;
    }
    
    // Restart scale
    double Q2restart = 0.0;
    // Strong ordering: Restart at evolution variable
    if (orderingMode == 1) Q2restart = Q2evol;
    
    // AlphaS and PDF scales
    // TODO VINCIA SHOWER alphaS has alphaSMax, how to include that?
    double Q2alphaS = max(mu2min, mu2freeze + kMu2*Q2evol/normFac);
    double Q2pdf    = Q2evol;
    
    if (PRINTNOW) {
      cout << " Q2evol    for " << name << " = " << Q2evol << endl;
      cout << " Q2restart for " << name << " = " << Q2restart << endl;
      cout << " Q2alphaS  for " << name << " = " << Q2alphaS << endl;
      cout << " Q2pdf     for " << name << " = " << Q2pdf << endl;
    }
    
    // Insert values in map
    evolVals.insert( pair<string, double>("t",        Q2evol)    );
    evolVals.insert( pair<string, double>("tRS",      Q2restart) );
    evolVals.insert( pair<string, double>("scaleAS",  Q2alphaS)  );
    evolVals.insert( pair<string, double>("scalePDF", Q2pdf)     );
    
    return evolVals;
  }*/
  
  // Check if attempted clustering is handled by timelike shower
  bool TimeAntShower::isTimelike( const Event& event, int iRa, int iEm, 
                                  int iRe, string ) {
    // Check if all have colours
    if ( event[iRa].colType() == 0 ||
         event[iRe].colType() == 0 ||
         event[iEm].colType() == 0 ) return false;
    return ( event[iRa].isFinal() && event[iEm].isFinal() 
                                  && event[iRe].isFinal() );
  }
  
  // Return a string identifier of a splitting.
  string TimeAntShower::getSplittingName( const Event& event, int iRa, 
                                          int iEm, int iRe ) { 
    if (!isTimelike(event, iRa, iEm, iRe, "")) return "";
    
    string name = "";
    
    int idEm = event[iEm].idAbs();
    
    // 1. Gluon emission
    if (idEm == 21) {
      int idRa = event[iRa].idAbs();
      int idRe = event[iRe].idAbs();
           if ( (idRa  <  7) && (idRe  <  7) ) name = "Vincia:QQemit";
      else if ( (idRa == 21) && (idRe == 21) ) name = "Vincia:GGemit";
      else if ( (idRa == 21) && (idRe  <  7) ) name = "Vincia:QGemit";
      else if ( (idRa  <  7) && (idRe == 21) ) name = "Vincia:QGemit";
    }
    
    // 2. Gluon splitting
    // emission id = - radiator id
    else if (idEm < 7) {
      if (event[iEm].id() == -event[iRa].id()) {
        int idRe = event[iRe].idAbs();
             if (idRe == 21) name = "Vincia:GGsplit";
        else if (idRe  <  7) name = "Vincia:QGsplit";
      }
    }
    
    if (PRINTNOW) {
      if (name != "") cout << " Splitting name = " << name << endl;
      else cout << " Did not find splitting" << endl;
    }
    
    return name;
  }
  
  // Return the splitting probability.
  double TimeAntShower::getSplittingProb( const Event& event, int iRa, 
                                          int iEm, int iRe, string name ) { 
    if (!isTimelike(event, iRa, iEm, iRe, name)) return 0.0;
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return 0.0;
    
    // MECs?
    int PMEnow = 1.0;
    if (meCorrections) {
      vector<Particle> partsForMEC;
      partsForMEC.clear();
      // quark line needed, Higgs-Interference, intermediate boson
      bool needQuarkLine = false;
      bool foundHiggs    = false;
      int imBoson        = 0;
      // EE
      if (incomingType == 1) {
        // Resonance if found, can only do Z and H here
        for (int i = 0; i < event.size(); ++i) if (!event[i].isFinal()) {
          if (event[i].idAbs() == 23) {
            partsForMEC.push_back(event[i]);
            needQuarkLine = true;
          } else if (event[i].idAbs() == 25) {
            partsForMEC.push_back(event[i]);
            foundHiggs = true;
          }
        }
        // Found too much -> can't continue
        if (partsForMEC.size() >= 2) {
          partsForMEC.clear();
          partsForMEC.resize(0);
        }
      }
      // PP
      else if (incomingType == 2) {
        // Incoming partons
        for (int i = 0; i < event.size(); ++i) if (!event[i].isFinal())
          if ( (event[i].mother1() == 1) || (event[i].mother1() == 2) )
            partsForMEC.push_back(event[i]);
        // Missing -> can't continue
        if (partsForMEC.size() < 2) {
          partsForMEC.clear();
          partsForMEC.resize(0);
        } else {
          // Check if we have intermediate bosons
          for (int i = 0; i < event.size(); ++i) if (!event[i].isFinal()) {
            if (event[i].idAbs() > 22 && event[i].idAbs() < 26) {
              partsForMEC.push_back(event[i]);
              imBoson = i;
              if (event[i].idAbs() == 25) foundHiggs = true;
              else needQuarkLine = true;
            }
          }
          // Found too much -> can't continue
          if (partsForMEC.size() >= 4) {
            partsForMEC.clear();
            partsForMEC.resize(0);
          } else {
            // Put boson as final
            partsForMEC[(int)partsForMEC.size()-1].statusNeg();
          }
        }
      }
      // Continue with FS particles, if we found incoming
      int nFScol = 0;
      if (partsForMEC.size() > 0) {
        for (int i=0; i < event.size(); i++) if (event[i].isFinal()) {
          if (event[i].colType() != 0) nFScol++;
          // Kick out decay products
          if (imBoson > 0 && event[i].colType() == 0) {
            if (event[imBoson].daughter1() == i) continue;
            if (event[imBoson].daughter2() == i) continue;
          }
          partsForMEC.push_back(event[i]);
        }
      }
      // Do have particles left that are not handled in the MECs
      for (int i = 0; i < (int)partsForMEC.size(); i++) {
        if (partsForMEC[i].idAbs() > 22 && partsForMEC[i].idAbs() > 26) continue;
        if (partsForMEC[i].colType() != 0) continue;
        partsForMEC.clear();
        partsForMEC.resize(0);
        break;
      }
      // Check the order
      bool canDoMECsNow = true;
      int MEclassNow    = 0;
      if (incomingType == 1) { // EE
        MEclassNow = singleBosonDecay;
        if ((nFScol-2) > matchingLO) canDoMECsNow = false;
      } else if (incomingType == 2) { // PP
        if (imBoson > 0) {
          MEclassNow = singleBosonProd;
          if (nFScol > matchingLO) canDoMECsNow = false;
        } else if (imBoson == 0) {
          MEclassNow = pureQCD;
          if ((nFScol-2) > matchingLO) canDoMECsNow = false;
        }
      }
      // Check if we can do MECs
      if (canDoMECsNow) {
        // For Higgs check interference
        int HintNow = 0;
        if (foundHiggs && !Hinterference) {
          if (MEclassNow == singleBosonProd) {
            // Check Higgs daughters
            int idDau1 = event[partsForMEC[0].daughter1()].id();
            int idDau2 = event[partsForMEC[0].daughter2()].id();
            HintNow = ( (idDau1 != 21 && idDau2 != 21) ? 1 : 2 );
          } else if (MEclassNow == singleBosonDecay) {
            // Check Higgs mothers
            int idMot1 = event[partsForMEC[2].mother1()].id();
            int idMot2 = event[partsForMEC[2].mother2()].id();
            HintNow = ( (idMot1 != 21 && idMot2 != 21) ? 1 : 2 );
          }
        }
        // Set properties
        MEcorrectionsPtr->setPari(false); // PARI TODO
        MEcorrectionsPtr->setProcClass(MEclassNow);
        MEcorrectionsPtr->setClustProps(needQuarkLine ? 1 : 0, HintNow);
        canDoMECsNow = MEcorrectionsPtr->canDoMEC(partsForMEC, true);
        PMEnow = MEcorrectionsPtr->MEcorrection(partsForMEC)[0];
      }
    }
    
    double splitProb = 0.0;
    
    // Find the right antenna
    for (int iant = iQQemit; iant <= iGGsplit; iant++) {
      if (name != getAnt(iant)->vinciaName()) continue;
      
      double sik    = (event[iRa].p()+event[iRe].p()).m2Calc();
      double sij    = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjk    = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sIK    = sik + sij + sjk;
      
      // sij and sjk need to be swapped for QGemit if Ra is gluon
      if (iant == iQGemit && event[iRa].id() == 21) {
        int tmp = sjk;
        sjk = sij;
        sij = tmp;
      }
      
      // col*ant
      double ant    = getAnt(iant)->antennaFunction(sij/sIK,sjk/sIK)/sIK;
      double colFac = getAnt(iant)->getChargeFactor();
      
      // CMW
      if (event[iEm].id() == 21) {
        if (CMWtype > 0) {
          double mu2  = 0.0;
          double kMu2 = 0.0;
          if (event[iEm].id() == 21) {
            mu2  = sij*sjk/sIK;
            kMu2 = aSkMu2;
          } else {
            mu2  = sij;
            kMu2 = aSkMu2Split;
          }
          mu2 = max(mu2min, mu2freeze + mu2*kMu2);
          double alphaSn = min(alphaSmax, alphaSptr->alphaS(mu2));
          double aSratio = min(alphaSmax, alphaSCMWptr->alphaS(mu2)) / alphaSn;
          double eik     = ( CMWtype == 1 ? 2.0*sIK/sij/sjk : 2.0*sik/sij/sjk );
          ant += (eik*(aSratio-1.0));
        }
        ant *= 0.5; // will get iRa <-> iRe combi
      }
      
      // Phase Space factor
      double PSfac = 1.0/sIK;
      
      // TODO
      // Antenna Mass Correction Terms
      // HYPERJET
      // PARI
      
      if (PRINTDETAILSNOW) {
        cout << " col*ant = " << colFac*ant << endl;
        cout << " PSfac   = " << PSfac << endl;
        cout << " PMEnow  = " << PMEnow << endl;
      }
      
      splitProb = PSfac*colFac*ant*PMEnow;
      break;
    }
    
    if (PRINTNOW) {
      cout << " Splitting probability for " << name << " = "
           << scientific << splitProb << endl;
      cout.precision(3); 
      cout << fixed;
    }
    
    return splitProb;
  }
  
  //**********************************************************************
  // SpaceAntShower
  //**********************************************************************
  
  // Return clustering kinematics - as needed for merging.
  Event SpaceAntShower::clustered( const Event& eventOld, int iRa, 
                                   int iEm, int iRe, string name ) {
    if (!isSpacelike(eventOld, iRa, iEm, iRe, name)) return Event();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return Event();
    
    Event event = eventOld;
    
    Particle radiator = event[iRa];
    Particle emission = event[iEm];
    Particle recoiler = event[iRe];
    
    // Check if II or IF
    bool isII = ( !radiator.isFinal() && !recoiler.isFinal() );
    
    // II - start
    if (isII) {
      
      double sab = (radiator.p()+recoiler.p()).m2Calc();
      double saj = (radiator.p()+emission.p()).m2Calc();
      double sjb = (emission.p()+recoiler.p()).m2Calc();
      double sAB = sab - saj - sjb;
      if (sAB < 0.0) return Event();
      
      // Clustering: kinematic part
      Vec4 pA = radiator.p();
      Vec4 pB = recoiler.p();
      double rescaleFacA = sqrt(sAB/sab*(sAB+saj)/(sAB+sjb));
      double rescaleFacB = sqrt(sAB/sab*(sAB+sjb)/(sAB+saj));
      pA.rescale4(rescaleFacA);
      pB.rescale4(rescaleFacB);
      // Check energys -> is clustering possible?
      if ( (pA.e() < 0.0) || (pA.e() > radiator.e()) ) return Event();
      if ( (pB.e() < 0.0) || (pB.e() > recoiler.e()) ) return Event();
    
      // Id's and colours
      int idA   = radiator.id();
      int colA  = radiator.col();
      int acolA = radiator.acol();
      int idB   = recoiler.id();
      int colB  = recoiler.col();
      int acolB = recoiler.acol();
      int idj   = emission.id();
      int colj  = emission.col();
      int acolj = emission.acol();
      // Gluon emission
      if (idj == 21) {
        if (colj == colB) colB = acolA;
        else if (acolj == acolB) acolB = colA;
      }
      // Initial gluon splitting
      else if (name == "Vincia:QXsplitII") {
        idA = -idj;
        if (idA > 0) acolA = 0;
        else if (idA < 0) colA = 0;
      }
      // Initial gluon conversion
      else if (name == "Vincia:GXconvII") {
        idA = 21;
        if (colA == 0) colA = acolj;
        if (acolA == 0) acolA = colj;
      } else return Event();
      
      // Replace radiator and recoiler in event record
      event[iRa].id(idA);
      event[iRa].p(pA);
      event[iRa].col(colA);
      event[iRa].acol(acolA);
      
      event[iRe].id(idB);
      event[iRe].p(pB);
      event[iRe].col(colB);
      event[iRe].acol(acolB);
      
      // Particle vector for FS/IM recoilers
      // TODO this only works if there is only one system in the event
      vector<Particle> FSrecoilers; 
      FSrecoilers.resize(0);
      vector<Particle> IMrecoilers;
      IMrecoilers.resize(0);
      for (int i=0; i<event.size(); i++) {
        // Final state
        if (event[i].isFinal() && (i != iEm)) FSrecoilers.push_back(event[i]);
        // Intermediate state, if present
        else if (event[i].statusAbs() == 22) IMrecoilers.push_back(event[i]);
      }
      // 1) Boost in center-of-mass frame AB, from E,0,0,pz to E,0,0,0
      Vec4 pSum = pA + pB;
      // 2) Boost from E,0,0,0 to E',px',py',pz' (FSrecoilers before clustering in lab frame)
      Vec4 pRec(0.0, 0.0, 0.0, 0.0);
      for (int i=0; i<(int)FSrecoilers.size(); i++) pRec += FSrecoilers[i].p();
      // Now we perform the boost back
      for (int i=0; i<(int)FSrecoilers.size(); i++) {
        FSrecoilers[i].bstback(pRec);
        FSrecoilers[i].bst(pSum);
      }
      for (int i=0; i<(int)IMrecoilers.size(); i++) {
        IMrecoilers[i].bstback(pRec);
        IMrecoilers[i].bst(pSum);
      }
      
      // Change momentum in event record
      int countIM = 0;
      int countFS = 0;
      for (int i=0; i<event.size(); i++) {
        if (event[i].isFinal() && (i != iEm)) {
          event[i].p(FSrecoilers[countFS].p());
          countFS++;
        }
        if (event[i].statusAbs() == 22) {
          event[i].p(IMrecoilers[countIM].p());
          countIM++;
        }
      }
      
    }
    // II - end
    
    // IF - start
    else {
      
      Particle isa = (radiator.isFinal() ? recoiler : radiator);
      Particle fsk = (radiator.isFinal() ? radiator : recoiler);
      double sak = (isa.p()+fsk.p()).m2Calc();
      double saj = (isa.p()+emission.p()).m2Calc();
      double sjk = (emission.p()+fsk.p()).m2Calc();
      double sAK = sak + saj - sjk;
      if (sAK < 0.0) return Event();
    
      // Clustering: kinematic part
      Vec4 pA = isa.p();
      pA.rescale4(sAK/(sAK+sjk));
      Vec4 pK = pA-isa.p()+fsk.p()+emission.p();
      // Check energys -> is clustering possible?
      if ( (pA.e() < 0.0) || (pK.e() < 0.0) ) return Event();
    
      // Id's and colours
      int idA    = isa.id();
      int colA   = isa.col();
      int acolA  = isa.acol();
      int idK    = fsk.id();
      int colK   = fsk.col();
      int acolK  = fsk.acol();
      int idj    = emission.id();
      int colj   = emission.col();
      int acolj  = emission.acol();
      // Gluon emission
      if (idj == 21) {
        if (colj == acolK) acolK = acolA;
        else if (acolj == colK) colK = colA;
      }
      // Initial gluon splitting
      else if (name == "Vincia:QXsplitIF") {
        idA = -idj;
        if (idA > 0) acolA = 0;
        else if (idA < 0) colA = 0;
      }
      // Initial gluon conversion
      else if (name == "Vincia:GXconvIF") {
        idA = 21;
        if (colA == 0) colA = acolj;
        if (acolA == 0) acolA = colj;
      }
      // Final gluon splitting
      else if (name == "Vincia:XGsplitIF") {
        idK = 21;
        if (colK == 0) colK = colj;
        if (acolK == 0) acolK = acolj;
      } else return Event();
      
      // Replace radiator and recoiler in event record
      if (event[iRa].isFinal()) {
        event[iRe].id(idA);
        event[iRe].p(pA);
        event[iRe].col(colA);
        event[iRe].acol(acolA);
        
        event[iRa].id(idK);
        event[iRa].p(pK);
        event[iRa].col(colK);
        event[iRa].acol(acolK);
      } else {
        event[iRa].id(idA);
        event[iRa].p(pA);
        event[iRa].col(colA);
        event[iRa].acol(acolA);
        
        event[iRe].id(idK);
        event[iRe].p(pK);
        event[iRe].col(colK);
        event[iRe].acol(acolK);
      }
      
    }
    // IF - end
      
    // Remove emission from event record
    event.remove(iEm, iEm);
    // Remove emission from mother/daughter lists
    for (int i=0; i<event.size(); i++) {
      // Mother1
      if (event[i].mother1() > iEm) 
        event[i].mother1(event[i].mother1()-1);
      else if (event[i].mother1() == iEm) {
        if (event[i].mother2() == iEm) 
          event[i].mothers(0, 0);
        else if (event[i].mother2() > iEm)
          event[i].mothers(event[i].mother2()-1, 0);
        else
          event[i].mothers(event[i].mother2(), 0);
      }
      // Mother2
      if (event[i].mother2() > iEm) 
        event[i].mother2(event[i].mother2()-1);
      else if (event[i].mother2() == iEm) 
        event[i].mother2(0);
      // Daugther1
      if (event[i].daughter1() > iEm) 
        event[i].daughter1(event[i].daughter1()-1);
      else if (event[i].daughter1() == iEm) {
        if (event[i].daughter2() == iEm)
          event[i].daughters(0, 0);
        else if (event[i].daughter2() > iEm)
          event[i].daughters(iEm, event[i].daughter2()-1);
        else
          event[i].daughters(event[i].daughter2(), 0);
      }
      // Daugther2
      if (event[i].daughter2() > iEm) 
        event[i].daughter2(event[i].daughter2()-1);
      else if (event[i].daughter2() == iEm) {
        if (event[i].daughter1() < iEm-1)
          event[i].daughters(event[i].daughter1(), iEm-1);
        else 
          event[i].daughter2(0);
      }
    }
    
    // Check energy-momentum conservation
    if (verbose >= 3)  {
      // Start with incoming partons = beam daughters
      Vec4 pTot = -(event[event[1].daughter1()].p() + 
                    event[event[2].daughter1()].p());
      double eLab = abs(pTot.e());
      // Sum final state
      for (int i=0; i<event.size(); ++i)
        if (event[i].isFinal()) pTot += event[i].p();
      // Check energy-momentum/charge conservation.
      double epDev = abs( pTot.e()) + abs(pTot.px()) + abs(pTot.py())
                                    + abs(pTot.pz() );
      if (epDev > vinComPtr->epTolErr * eLab) {
        cout << "ERROR in SpaceAntShower::clustered for " << name
             << ": energy-momentum not conserved" << endl;
        for (int i=0; i<(int)eventOld.size(); i++) if (event[i].isFinal())
          cout << event[i].m2Calc() << endl;
        cout << " Unclustered event " << name << ":" << endl;
        eventOld.list();
        cout << " Clustered event for " << name << ":" << endl;
        event.list();
      }
      else if (epDev > vinComPtr->epTolWarn * eLab) {
        cout << "WARNING in SpaceAntShower::clustered for " << name
             << ": energy-momentum not quite conserved" << endl;
      }
    }
    
    if (PRINTNOW) {
      cout << " Unclustered event for " << name << ":" << endl;
      eventOld.list();
      cout << " Clustered event for " << name << ":" << endl;
      event.list();
    }
    
    return event;
  }
  
  // Return the evolution variable.
  // Important note:
  // - This map must contain an entry for the shower evolution variable, 
  //   specified with key "t".
  // - This map must contain an entry for the shower evolution variable from
  //   which the shower would be restarted after a branching. This entry
  //   must have key "tRS", 
  // - This map must contain an entry for the argument of \alpha_s used
  //   for the branching. This entry must have key "scaleAS". 
  // - This map must contain an entry for the argument of the PDFs used
  //   for the branching. This entry must have key "scalePDF". 
  /*map<string, double> SpaceAntShower::getStateVariables( const Event& event, int iRa, 
                                                         int iEm, int iRe, string name) {
    map<string, double> evolVals;
    evolVals.clear();
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return evolVals;
    
    // Erase the Vincia: and II/IF part to minimize string checks
    name.erase(name.begin(), name.begin()+vincia.size());
    name.erase(name.end()-2, name.end());
    
    bool isII = ( !event[iRa].isFinal() && !event[iRe].isFinal() );
    int idEm = event[iEm].idAbs();
    
    // Evolution variable
    double Q2evol  = 0.0;
    double kMu2    = 0.0;
    double normFac = 1.0;
    
    // 1. Gluon emission:
    if (idEm == 21) {
      kMu2 = aSkMu2;
      // 1.1 from II, symmetric
      if (isII) {
        Q2evol  = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 1);
        normFac = evNormIIpT2;
      } else {
        normFac = evNormIFpT2;
        // 1.2 from IF side1
        if (!event[iRa].isFinal())
          Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 1);
        // 1.2 from IF side2
        else if (!event[iRe].isFinal())
          Q2evol = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 1);
      }
    }
    
    // Non-emission
    else if (idEm < 7) {
      
      // 2. Initial gluon splitting: saj
      if (name == "QXsplit") {
        kMu2    = aSkMu2SplitI;
        normFac = evNormMsplitI;
        // 2.1 from II
        if (isII) {
          Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 2 : 1);
          if (evAllPT) normFac = evNormIIpT2;
        }
        // 2.2 form IF
        else {
          Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 2 : 1);
          if (evAllPT) normFac = evNormIFpT2;
        }
      }
      
      // 3. Gluon conversion: saj
      else if (name == "GXconv") {
        kMu2    = aSkMu2Conv;
        normFac = evNormMconvI;
        // 3.1 from II
        if (isII) {
          Q2evol = resolutionPtr->Q2EII(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 2 : 1);
          if (evAllPT) normFac = evNormIIpT2;
        } 
        // 3.2 form IF
        else {
          Q2evol = resolutionPtr->Q2EIF(event[iRa], event[iRe], event[iEm], 
                                        !evAllPT ? 3 : 1);
          if (evAllPT) normFac = evNormIFpT2;
        }
      }
      
      // 4. Final gluon splitting: sjk
      else if (name == "XGsplit") {
        Q2evol = resolutionPtr->Q2EIF(event[iRe], event[iRa], event[iEm], 
                                      !evAllPT ? 4 : 1);
        kMu2    = aSkMu2SplitF;
        normFac = (evAllPT ? evNormIFpT2 : evNormMsplitF);
      }
    }
    
    // Restart scale
    double Q2restart = 0.0;
    // Strong ordering: Restart at evolution variable
    if (orderingMode == 1) Q2restart = Q2evol;
    
    // AlphaS and PDF scales
    // TODO VINCIA SHOWER alphaS has alphaSMax, how to include that?
    double Q2alphaS = max(mu2min, mu2freeze + kMu2*Q2evol/normFac);
    double Q2pdf    = Q2evol;
    
    if (PRINTNOW) {
      cout << " Q2evol    for " << name << " = " << Q2evol << endl;
      cout << " Q2restart for " << name << " = " << Q2restart << endl;
      cout << " Q2alphaS  for " << name << " = " << Q2alphaS << endl;
      cout << " Q2pdf     for " << name << " = " << Q2pdf << endl;
    }
    
    // Insert values in map
    evolVals.insert( pair<string, double>("t",        Q2evol)    );
    evolVals.insert( pair<string, double>("tRS",      Q2restart) );
    evolVals.insert( pair<string, double>("scaleAS",  Q2alphaS)  );
    evolVals.insert( pair<string, double>("scalePDF", Q2pdf)     );
    
    return evolVals;
  }*/
  
  // Check if attempted clustering is handled by spacelike shower
  bool SpaceAntShower::isSpacelike( const Event& event, int iRa, int iEm, 
                                    int iRe, string ) {
    // Check if all have colours
    if ( event[iRa].colType() == 0 ||
         event[iRe].colType() == 0 ||
         event[iEm].colType() == 0 ) return false;
    return !( event[iRa].isFinal() && event[iEm].isFinal() 
                                   && event[iRe].isFinal() );
  }
  
  // Return a string identifier of a splitting.
  string SpaceAntShower::getSplittingName( const Event& event, int iRa, 
                                           int iEm, int iRe ) { 
    if (!isSpacelike(event, iRa, iEm, iRe, "")) return "";
    
    string name = "";
    
    bool iRaIS = !event[iRa].isFinal();
    bool iReIS = !event[iRe].isFinal();
    bool isII  = (iRaIS && iReIS);
    int  idRa  = event[iRa].idAbs();
    int  idEm  = event[iEm].idAbs();
    int  idRe  = event[iRe].idAbs();
    
    // 1. Gluon emission
    if (idEm == 21) {
      // 1.1 from II
      if (isII) {
             if ( (idRa  <  7) && (idRe  <  7) ) name = "Vincia:QQemitII";
        else if ( (idRa == 21) && (idRe == 21) ) name = "Vincia:GGemitII";
        else if ( (idRa == 21) && (idRe  <  7) ) name = "Vincia:GQemitII";
        else if ( (idRa  <  7) && (idRe == 21) ) name = "Vincia:GQemitII";
      }
      // 1.2 from IF
      else if (iRaIS || iReIS) {
        int idIS = (!iReIS ? idRa : idRe);
        int idFS = (!iReIS ? idRe : idRa);
             if ( (idIS  <  7) && (idFS  <  7) ) name = "Vincia:QQemitIF";
        else if ( (idIS == 21) && (idFS == 21) ) name = "Vincia:GGemitIF";
        else if ( (idIS == 21) && (idFS  <  7) ) name = "Vincia:GQemitIF";
        else if ( (idIS  <  7) && (idFS == 21) ) name = "Vincia:QGemitIF";
      }
    }
    
    // Non-emission
    else if (idEm < 7) {
      
      // 2. Initial gluon splitting
      // radiator = initial gluon, radiator in initial state
      if ( (idRa == 21) && iRaIS ) {
        // 2.1 from II
        if (isII) name = "Vincia:QXsplitII";
        // 2.2 form IF
        else      name = "Vincia:QXsplitIF";
      }
      
      // 3. Gluon conversion
      // radiator id = emission id, radiator in initial state
      else if ( (event[iEm].id() == event[iRa].id()) && iRaIS ) {
        // 2.1 from II
        if (isII) name = "Vincia:GXconvII";
        // 2.2 form IF
        else      name = "Vincia:GXconvIF";
      }
      
      // 4. Final gluon splitting
      // emission id = - radiator id, radiator in final state
      else if ( (event[iEm].id() == -event[iRa].id()) && !iRaIS ) {
        name = "Vincia:XGsplitIF";
      }
      
    }
    
    if (PRINTNOW) {
      if (name != "") cout << " Splitting name = " << name << endl;
      else cout << " Did not find splitting" << endl;
    }
    
    return name;
  }
  
  // Return the splitting probability.
  double SpaceAntShower::getSplittingProb( const Event& event, int iRa, 
                                           int iEm, int iRe, string name ) { 
    if (!isSpacelike(event, iRa, iEm, iRe, name)) return 0.0;
    string vincia("Vincia:");
    if (name.find(vincia) == string::npos) return 0.0;
    
    // MECs?
    int PMEnow = 1.0;
    if (/*meCorrections*/false) {
      vector<Particle> partsForMEC;
      partsForMEC.clear();
      // quark line needed, Higgs-Interference, intermediate boson
      bool needQuarkLine = false;
      bool foundHiggs    = false;
      int imBoson        = 0;
      // no EE
      // PP
      if (incomingType == 2) {
        // Incoming partons
        for (int i = 0; i < event.size(); ++i) if (!event[i].isFinal())
          if ( (event[i].mother1() == 1) || (event[i].mother1() == 2) )
            partsForMEC.push_back(event[i]);
        // Missing -> can't continue
        if (partsForMEC.size() < 2) {
          partsForMEC.clear();
          partsForMEC.resize(0);
        } else {
          // Check if we have intermediate bosons
          for (int i = 0; i < event.size(); ++i) if (!event[i].isFinal()) {
            if (event[i].idAbs() > 22 && event[i].idAbs() < 26) {
              partsForMEC.push_back(event[i]);
              imBoson = i;
              if (event[i].idAbs() == 25) foundHiggs = true;
              else needQuarkLine = true;
            }
          }
          // Found too much -> can't continue
          if (partsForMEC.size() >= 4) {
            partsForMEC.clear();
            partsForMEC.resize(0);
          } else {
            // Put boson as final
            partsForMEC[(int)partsForMEC.size()-1].statusNeg();
          }
        }
      }
      // Continue with FS particles, if we found incoming
      int nFScol = 0;
      if (partsForMEC.size() > 0) {
        for (int i=0; i < event.size(); i++) if (event[i].isFinal()) {
          if (event[i].colType() != 0) nFScol++;
          // Kick out decay products
          if (imBoson > 0 && event[i].colType() == 0) {
            if (event[imBoson].daughter1() == i) continue;
            if (event[imBoson].daughter2() == i) continue;
          }
          partsForMEC.push_back(event[i]);
        }
      }
      // Do have particles left that are not handled in the MECs
      for (int i = 0; i < (int)partsForMEC.size(); i++) {
        if (partsForMEC[i].idAbs() > 22 && partsForMEC[i].idAbs() > 26) continue;
        if (partsForMEC[i].colType() != 0) continue;
        partsForMEC.clear();
        partsForMEC.resize(0);
        break;
      }
      // Check the order
      bool canDoMECsNow = true;
      int MEclassNow    = 0;
      if (incomingType == 2) { // PP
        if (imBoson > 0) {
          MEclassNow = singleBosonProd;
          if (nFScol > matchingLO) canDoMECsNow = false;
        } else if (imBoson == 0) {
          MEclassNow = pureQCD;
          if ((nFScol-2) > matchingLO) canDoMECsNow = false;
        }
      }
      // Check if we can do MECs
      if (canDoMECsNow) {
        // For Higgs check interference
        int HintNow = 0;
        if (foundHiggs && !Hinterference) {
          if (MEclassNow == singleBosonProd) {
            // Check Higgs daughters
            int idDau1 = event[partsForMEC[0].daughter1()].id();
            int idDau2 = event[partsForMEC[0].daughter2()].id();
            HintNow = ( (idDau1 != 21 && idDau2 != 21) ? 1 : 2 );
          } else if (MEclassNow == singleBosonDecay) {
            // Check Higgs mothers
            int idMot1 = event[partsForMEC[2].mother1()].id();
            int idMot2 = event[partsForMEC[2].mother2()].id();
            HintNow = ( (idMot1 != 21 && idMot2 != 21) ? 1 : 2 );
          }
        }
        // Set properties
        MEcorrectionsPtr->setPari(false); // PARI TODO
        MEcorrectionsPtr->setProcClass(MEclassNow);
        MEcorrectionsPtr->setClustProps(needQuarkLine ? 1 : 0, HintNow);
        canDoMECsNow = MEcorrectionsPtr->canDoMEC(partsForMEC, true);
        PMEnow = MEcorrectionsPtr->MEcorrection(partsForMEC)[0];
      }
    }
    
    double splitProb = 0.0;
    
    bool isII = ( !event[iRa].isFinal() && !event[iRe].isFinal() );
    
    // Find the right antenna for II
    if (isII) { for (int iant = iQQemitII; iant <= iGXconvII; iant++) {
      if (name != getAnt(iant)->vinciaName()) continue;
      
      // Swap iRa and iRe for GQemitII if Re is gluon
      //                  for iQXsplitII if Em is not cc to Ra
      //                  for iGXconvII if id(Em) is not id(Ra)
      int colEm = (event[iEm].id() > 0 ? event[iEm].col() : event[iEm].acol());
      int colRa = (event[iEm].id() > 0 ? event[iRa].col() : event[iRa].acol());
      if ( (iant == iGQemitII  && event[iRe].id() == 21) ||
           (iant == iQXsplitII && colEm != colRa) ||
           (iant == iGXconvII  && event[iEm].id() != event[iRa].id()) ) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }
      
      double sab = (event[iRa].p()+event[iRe].p()).m2Calc();
      double saj = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjb = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sAB = (sab - saj - sjb);
      if (sAB < 0.0) return splitProb;
      
      // col*ant
      double ant    = getAnt(iant)->antennaFunction(saj,sjb,sAB);
      double colFac = getAnt(iant)->getChargeFactor();
      
      // CMW
      if (event[iEm].id() == 21) {
        if (CMWtypeII > 0) {
          double mu2  = 0.0;
          double kMu2 = 0.0;
          if (event[iEm].id() == 21) {
            mu2  = saj*sjb/sab;
            kMu2 = aSkMu2;
          } else if (iant == iQXsplitII) {
            mu2  = (evAllPT ? saj*sjb/sab : saj);
            kMu2 = aSkMu2SplitI;
          } else if (iant == iGXconvII) {
            mu2   = (evAllPT ? saj*sjb/sab : saj);
            kMu2 = aSkMu2Conv;
          }
          mu2 = max(mu2min, mu2freeze + mu2*kMu2);
          double alphaSn = min(alphaSmax, alphaSptr->alphaS(mu2));
          double aSratio = min(alphaSmax, alphaSCMWptr->alphaS(mu2)) / alphaSn;
          double eik     = ( CMWtypeII == 1 ? 2.0*sAB/saj/sjb : 2.0*sab/saj/sjb );
          ant += (eik*(aSratio-1.0));
        }
        ant *= 0.5; // will get iRa <-> iRe combi
      }
      
      // Phase Space factor
      double PSfac = sAB/sab/sab;
      
      if (PRINTDETAILSNOW) {
        cout << " col*ant = " << colFac*ant << endl;
        cout << " PSfac   = " << PSfac << endl;
        cout << " PMEnow  = " << PMEnow << endl;
      }
      
      splitProb = PSfac*colFac*ant*PMEnow;
      break;
    } }
    
    // Find the right antenna for IF
    else { for (int iant = iQQemitIF; iant <= iXGsplitIF; iant++) {
      // Find the right antenna
      if (name != getAnt(iant)->vinciaName()) continue;
      
      // Swap iRa and iRe if iRe is initial
      if (!event[iRe].isFinal()) {
        int tmp = iRa;
        iRa = iRe;
        iRe = tmp;
      }
      
      double sak = (event[iRa].p()+event[iRe].p()).m2Calc();
      double saj = (event[iRa].p()+event[iEm].p()).m2Calc();
      double sjk = (event[iEm].p()+event[iRe].p()).m2Calc();
      double sAK = (sak + saj - sjk);
      if (sAK < 0.0) return splitProb;
      
      // col*ant
      double ant    = getAnt(iant)->antennaFunction(saj,sjk,sAK);
      double colFac = getAnt(iant)->getChargeFactor();
      
      // CMW
      if (event[iEm].id() == 21) {
        if (CMWtypeIF > 0) {
          double mu2  = 0.0;
          double kMu2 = 0.0;
          if (event[iEm].id() == 21) {
            mu2  = saj*sjk/(sAK+sjk);
            kMu2 = aSkMu2;
          } else if (iant == iXGsplitIF) {
            mu2  = (evAllPT ? saj*sjk/(sAK+sjk) : sjk);
            kMu2 = aSkMu2SplitF;
          } else if (iant == iQXsplitIF) {
            mu2  = (evAllPT ? saj*sjk/(sAK+sjk) : saj);
            kMu2 = aSkMu2SplitI;
          } else if (iant == iGXconvIF) {
            mu2  = (evAllPT ? saj*sjk/(sAK+sjk) : saj);
            kMu2 = aSkMu2Conv;
          }
          mu2 = max(mu2min, mu2freeze + mu2*kMu2);
          double alphaSn = min(alphaSmax, alphaSptr->alphaS(mu2));
          double aSratio = min(alphaSmax, alphaSCMWptr->alphaS(mu2)) / alphaSn;
          double eik     = ( CMWtypeIF == 1 ? 2.0*sAK/saj/sjk : 2.0*sak/saj/sjk );
          ant += (eik*(aSratio-1.0));
        }
        ant *= 0.5; // will get iRa <-> iRe combi
      }
      
      // Phase Space factor
      double PSfac = sAK/(sAK+sjk);
      
      if (PRINTDETAILSNOW) {
        cout << " col*ant = " << colFac*ant << endl;
        cout << " PSfac   = " << PSfac << endl;
        cout << " PMEnow  = " << PMEnow << endl;
      }
      
      splitProb = PSfac*colFac*ant*PMEnow;
      break;
    } }
    
    if (PRINTNOW) {
      cout << " Splitting probability for " << name << " = "
           << scientific << splitProb << endl;
      cout.precision(3); 
      cout << fixed;
    }
    
    return splitProb;
  }

} // end namespace VINCIA
